import subprocess


def wslpath_convert(path: str, to_windows: bool = False) -> str:
    """Converts paths between WSL and Windows formats using wslpath."""
    command = ["wslpath", "-w" if to_windows else "-u", path]
    result = subprocess.run(command, capture_output=True, text=True, check=True)
    return result.stdout.strip()


def check_existing_outputs(files):
    """Check if all specified files exist."""
    return all(f.exists() for f in files)



import logging
from functools import partial
from pathlib import Path
import os
import glob
import multiprocessing
import subprocess
from rich.progress import Progress  # Replace tqdm with rich progress
from rich.logging import RichHandler  # Add rich log formatting
from b3p.ccx import mesh2ccx, ccx2vtu, ccxpost
from b3p.ccx.failcrit_mesh import compute_failure_for_meshes
from b3p.cli.app_state import AppState

logger = logging.getLogger(__name__)
logging.basicConfig(handlers=[RichHandler(rich_tracebacks=True)], level=logging.INFO)
# Configure rich for logging


def run_ccx(inp, ccxexe, logger):
    """Run CalculiX (ccx) on a given input file, capturing stdout/stderr."""
    cmd = [ccxexe, inp.replace(".inp", "")]
    logger.debug(f"Running command: {' '.join(cmd)}")
    try:
        result = subprocess.run(cmd, capture_output=True, text=True, check=True)
        logger.debug(f"CCX output for {inp}:\n{result.stdout}")
        return inp, True, None
    except subprocess.CalledProcessError as e:
        error_msg = (
            f"ccx failed for {inp}: {e}\nStdout:\n{e.stdout}\nStderr:\n{e.stderr}"
        )
        logger.error(error_msg)
        return inp, False, error_msg


def check_ccx_run_done(inpfile):
    """Check if the ccx run is done by looking for the .frd file."""
    frd_file = inpfile.replace(".inp", ".frd")
    if os.path.exists(frd_file):
        with open(frd_file, "rb") as f:
            f.seek(-5, 2)
            y = f.read()
            if y == b"9999\n":
                return True
    return False


class CcxApp:
    def __init__(self, state: AppState, yml: Path):
        self.state = state
        self.yml = yml
        self.dir = "fea"
        self.state.load_yaml(self.yml)

    def ccx(self, **kwargs):
        self.prep(**kwargs)
        self.solve(**kwargs)
        self.post(**kwargs)
        self.plot(**kwargs)

    def prep(self, bondline=False, **kwargs):
        base_prefix = self.state.get_prefix("drape")
        prefix = self.state.get_prefix(self.dir)
        available_meshes = glob.glob(f"{base_prefix}_joined.vtu")
        logger.info(f" available meshes {available_meshes}")
        if bondline:
            bondline_meshes = glob.glob(f"{base_prefix}*_bondline.vtu")
            if bondline_meshes:
                available_meshes = bondline_meshes

        logger.info(f"Available meshes: {available_meshes}")
        if not available_meshes:
            logger.error("No meshes found, did you build the blade geometry?")
            return

        output_files = mesh2ccx.mesh2ccx(
            available_meshes[-1],
            matmap=str(Path(base_prefix).parent / "material_map.json"),
            out=f"{prefix}_ccx.inp",
            bondline=bondline,
            **{k: v for k, v in kwargs.items() if k != "bondline"},
        )
        logger.info(f"Written: {', '.join(output_files)}")

    def solve(
        self,
        wildcard="",
        nproc=2,
        ccxexe="ccx",
        inpfiles=None,
        merged_plies=False,
        bondline=False,
        **kwargs,
    ):
        self.state.load_yaml(self.yml)
        prefix = self.state.get_prefix(self.dir)
        if inpfiles is None:
            inpfiles = glob.glob(f"{prefix}*ccx*{wildcard}*.inp")
        inps = [inp for inp in inpfiles]

        inps_to_run = [inp for inp in inps if not check_ccx_run_done(inp)]

        if not inps:
            logger.error(f"No input files found matching {prefix}*{wildcard}*inp")
            return
        else:
            logger.info(
                f"Found {len(inps)} input files matching {prefix}*{wildcard}*inp"
            )

        if inps_to_run:
            with multiprocessing.Pool(nproc) as pool:
                with Progress() as progress:  # Replace tqdm with rich Progress
                    task = progress.add_task("Running CCX", total=len(inps_to_run))
                    for inp, success, error_msg in pool.imap_unordered(
                        partial(run_ccx, ccxexe=ccxexe, logger=logger), inps_to_run
                    ):
                        progress.update(task, advance=1)
                        if not success:
                            logger.error(error_msg)

    def post(self, wildcard="", nbins=60, bondline=False, **kwargs):
        self.state.load_yaml(self.yml)
        prefix = self.state.get_prefix(self.dir)
        ccxpost = ccx2vtu.ccx2vtu(prefix, wildcard=wildcard)
        ccxpost.load_grids()
        ccxpost.tabulate(nbins)

    def failure_criteria(self, **kwargs):
        self.state.load_yaml(self.yml)
        prefix = self.state.get_prefix(self.dir)
        vtus = [i for i in prefix.parent.glob("*ccx*vtu") if str(i).find("fail") == -1]
        logger.info(f"VTU files found for failure criteria: {vtus}")

        compute_failure_for_meshes(vtus)

    def plot(self, wildcard="", plot3d=True, plot2d=True, bondline=False, **kwargs):
        self.state.load_yaml(self.yml)
        plotter = ccxpost.plot_ccx(self.state.get_workdir(), wildcard=wildcard)
        if plot3d:
            plotter.plot3d()
        if plot2d:
            plotter.plot2d(wildcard=wildcard)



# import os
# import shutil
# import pytest


# @pytest.fixture
# def prep_temp_dir(tmp_path_factory):
#     temp_dir = tmp_path_factory.mktemp("test_files")
#     example_dir = os.path.join(os.path.dirname(__file__), "..", "examples")
#     for item in os.listdir(example_dir):
#         s = os.path.join(example_dir, item)
#         d = os.path.join(temp_dir, item)
#         if os.path.isfile(s):
#             shutil.copy2(s, d)
#         else:
#             shutil.copytree(s, d)
#     yield temp_dir
#     shutil.rmtree(temp_dir)



import logging
import pandas as pd
import numpy as np
from copy import deepcopy as dc
from matplotlib import pyplot as plt
import pickle
import pyvista as pv
import json
from b3p.geometry import splining
from b3p.geometry import loft_utils
from b3p.geometry import blade_section

# Configure logging
logger = logging.getLogger(__name__)


class blade:
    def __init__(
        self,
        chord,
        thickness,
        twist,
        dx,
        dy,
        z,
        airfoils,
        chordwise_sampling,
        np_spanwise=100,
    ):
        self.np_spanwise = np_spanwise
        self.np_chordwise = len(chordwise_sampling)
        self._load_airfoils(airfoils, chordwise_sampling)
        self._interpolate_planform(chord, thickness, twist, dx, dy, z)
        self._place_airfoils()

    def to_table(self, x, prefix="prebend_out"):
        cols = [
            "relative_r",
            "z",
            "prebend",
            "chord",
            "relative_thickness",
            "absolute_thickness",
            "twist",
        ]
        df = pd.DataFrame(
            np.array(
                [x]
                + [
                    np.interp(x, i[0], i[1])
                    for i in [
                        self.z,
                        self.dx,
                        self.chord,
                        self.thickness,
                        self.absolute_thickness,
                        self.twist,
                    ]
                ]
            ).T,
            columns=cols,
        )
        df.to_csv(f"{prefix}.csv", index=False, sep=";")

    def _load_airfoils(self, airfoils, x):
        logger.info("Loading airfoils")
        self.airfoils = {}
        for i in sorted(airfoils):
            if type(airfoils[i]) == str:
                if airfoils[i].find("du") != -1:
                    logger.debug(f"Loading {airfoils[i]} normalized")
                    t = loft_utils.load(airfoils[i], normalise=True)
                else:
                    logger.debug(f"Loading {airfoils[i]} unnormalized")
                    t = loft_utils.load(airfoils[i], normalise=False)
            else:
                logger.debug(f"Loading airfoil {airfoils[i]['name']} at thickness {i}")
                t = airfoils[i]["xy"]

            self.airfoils[i] = loft_utils.interp(x, t)[:2]

    def _interpolate_planform(self, chord, thickness, twist, dx, dy, z):
        self.x = np.linspace(0, 1.0, self.np_spanwise)
        (
            self.input_chord,
            self.input_thickness,
            self.input_twist,
            self.input_dx,
            self.input_dy,
        ) = (
            list(zip(*dc(chord))),
            list(zip(*dc(thickness))),
            list(zip(*dc(twist))),
            list(zip(*dc(dx))),
            list(zip(*dc(dy))),
        )

        self.chord = splining.intp_c(self.x, chord)
        self.twist = splining.intp_c(self.x, twist)
        self.thickness = splining.intp_c(self.x, thickness)
        self.dx = splining.intp_c(self.x, dx)
        self.dy = splining.intp_c(self.x, dy)
        self.z = splining.intp_c(self.x, z)
        self.absolute_thickness = [
            self.x,
            [i[0] * i[1] for i in zip(self.chord[1], self.thickness[1])],
        ]

    def plot(self, name="_dum", fname="_out.png"):
        fig, ax = plt.subplots(3, 3, figsize=(15, 15))

        ax[0, 0].plot(self.x, self.chord[1], label=name)
        ax[0, 0].plot(
            self.input_chord[0], self.input_chord[1], "o", label=f"{name}_input"
        )
        ax[0, 0].set_xlabel("rel span (-)")
        ax[0, 0].legend(loc="best").get_frame().set_alpha(0.5)

        ax[2, 1].plot(self.x, self.chord[1], label=name)
        ax[2, 1].plot(
            self.input_chord[0], self.input_chord[1], "o", label=f"{name}_input"
        )
        ax[2, 1].grid(True)
        ax[2, 1].set_xlim(0.9, 1)

        ax[0, 1].plot(self.x, self.twist[1], label=name)
        ax[0, 1].plot(
            self.input_twist[0], self.input_twist[1], "o", label=f"{name}_input"
        )
        ax[0, 2].plot(self.x, self.thickness[1], label=name)
        ax[0, 2].plot(
            self.input_thickness[0],
            self.input_thickness[1],
            "o",
            label=f"{name}_input",
        )
        ax[1, 0].plot(
            self.absolute_thickness[0], self.absolute_thickness[1], label=name
        )
        ax[2, 0].plot(
            self.absolute_thickness[0], self.absolute_thickness[1], label=name
        )
        ax[2, 0].set_xlim(0, 0.2)

        ax[1, 1].plot(self.dx[0], self.dx[1], label=f"{name}_x")
        ax[1, 1].plot(self.input_dx[0], self.input_dx[1], "o", label=f"{name}_input")
        ax[1, 2].plot(self.dy[0], self.dy[1], label=f"{name}_y")
        ax[1, 2].plot(self.input_dy[0], self.input_dy[1], "o", label=f"{name}_input")

        titles = [
            "chord rotor_diam=%.3f" % (2.0 * max(self.z[1])),
            "twist",
            "relative thickness",
            "absolute thickness",
            "dx",
            "dy",
            "abs thickness root",
            "tip chord",
            "",
        ]

        for i in zip(ax.flatten(), titles):
            i[0].grid(True)
            i[0].set_title(i[1])

        ax[2, 2].remove()

        fig.tight_layout()
        fig.savefig(fname, dpi=100)

    def _interpolate_airfoils(self):
        v = [
            list(
                zip(
                    [i for _ in self.airfoils[i][0]],
                    self.airfoils[i][0],
                    self.airfoils[i][1],
                )
            )
            for i in sorted(self.airfoils)
        ]
        nv = []
        for i in zip(*v):
            t, x, y = zip(*i)
            nx = np.interp(self.thickness[1], t, x)
            ny = np.interp(self.thickness[1], t, y)
            nv.append(list(zip(nx, ny)))

        sections = []
        for i in zip(*nv):
            x, y = list(zip(*i))
            sections.append(blade_section.section(x, y))

        return sections

    def _place_airfoils(self):
        self.sections = self._interpolate_airfoils()

        twist_center = 0.5

        for i in zip(self.x, self.chord[1], self.twist[1], self.sections):
            i[3].translate(-twist_center, 0.0, 0.0)
            i[3].twist(i[2])
            i[3].scale((i[1], i[1], 1.0))

        for i in self.sections:
            i.local_to_global()

        for i in zip(self.sections, self.dx[1], self.dy[1], self.z[1]):
            i[0].translate(i[1], i[2], i[3])

    def export_variables(self, fname):
        var = {
            "dx": self.dx,
            "dy": self.dy,
            "z": self.z,
            "twist": self.twist,
            "chord": self.chord,
            "thickness": self.thickness,
            "absolute_thickness": self.absolute_thickness,
        }
        vv = {}
        for i in var:
            vv[i] = np.array(var[i]).tolist()

        json.dump(vv, open(fname, "w"))
        logger.info(f"Saved variables to {fname}")
        return var

    def dump(self, fname="__sections.txt", z_rotation=0.0):
        lst = [i.get_pointlist(z_rotation=z_rotation) for i in self.sections]
        if fname.endswith(".txt"):
            open(fname, "wb").write(str(lst).encode("utf-8"))
        elif fname.endswith(".pck"):
            pickle.dump(lst, open(fname, "wb"))

    def export_xfoil(self, prefix="airfoil_out/_xf"):
        for i in zip(self.sections, self.thickness[1], self.z[1]):
            nm = prefix + "_t_%.3f_r_%.3f" % (i[1], i[2])
            i[0].to_xfoil(nm.replace(".", "_") + ".dat")

    def mesh(self, fname=None):
        n_points = self.np_chordwise
        points = []
        for i in self.sections:
            for j in range(i.polydata.GetNumberOfPoints()):
                pt = i.polydata.GetPoint(j)
                points.append([pt[0], pt[1], pt[2]])

        points = np.array(points)

        cells = []
        for i in range(1, len(self.sections)):
            s0 = range((i - 1) * n_points, i * n_points)
            s1 = range(i * n_points, (i + 1) * n_points)
            cells.extend(
                [4, s0[j], s1[j], s1[(j + 1) % n_points], s0[(j + 1) % n_points]]
                for j in range(n_points)
            )

        self.poly = pv.PolyData(points, np.hstack(cells))
        if fname is not None:
            logger.info(f"Saving mesh to {fname}")
            self.poly.save(fname)



import pytest
import logging
from pathlib import Path
import shutil
import os
from unittest.mock import patch
from b3p.cli.app_state import AppState
from b3p.cli.clean_app import CleanApp
from b3p.cli.build_app import BuildApp
from b3p.cli.two_d_app import TwoDApp
from b3p.cli.ccx_app import CcxApp
from b3p.cli.utils import check_existing_outputs

# Configure logging for tests
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


@pytest.fixture
def temp_workdir(tmp_path):
    """Set up a temporary working directory with blade_test.yml."""
    files = [
        Path("examples/blade_test.yml"),
        Path("examples/materials_iso.yml"),
        Path("examples/loads_v0.yml"),
        Path("examples/laminates_v0.yml"),
    ]
    airfoils = Path("examples/airfoils")

    for src in files:
        if not src.exists():
            pytest.skip(f"{src} not found in examples/")
        dest = tmp_path / src.name
        shutil.copy(src, dest)
    shutil.copytree(airfoils, tmp_path / "airfoils", dirs_exist_ok=True)

    # Set CWD to tmp_path
    os.chdir(tmp_path)

    yield tmp_path, Path("blade_test.yml")
    # Cleanup
    state = AppState.get_instance()
    state.reset()


@pytest.fixture
def app_state():
    """Fixture for AppState singleton."""
    state = AppState.get_instance()
    yield state
    state.reset()


def test_clean(temp_workdir, app_state, caplog):
    """Test cleaning the working directory."""
    tmp_path, yml_path = temp_workdir
    caplog.set_level(logging.INFO)

    # Create a dummy workdir
    workdir = tmp_path / "temp_blade"
    workdir.mkdir()
    (workdir / "dummy.txt").write_text("test")

    clean_app = CleanApp(app_state, yml_path)
    clean_app.clean()

    assert not workdir.exists()
    assert "Removed workdir" in caplog.text


def test_build(temp_workdir, app_state, caplog):
    """Test the build process, skipping if outputs exist."""
    tmp_path, yml_path = temp_workdir
    caplog.set_level(logging.INFO)

    build_app = BuildApp(app_state, yml_path)

    # Check if key outputs already exist
    prefix = app_state.load_yaml(yml_path)["general"].get("prefix", "b3p")
    workdir = app_state.get_workdir()
    expected_files = [
        workdir / "drape" / f"{prefix}_joined.vtu",
        workdir / "drape" / f"{prefix}_mass.csv",
        workdir / "drape" / f"{prefix}_loads.png",
    ]

    if check_existing_outputs(expected_files):
        logger.info("Build outputs already exist, skipping build")
    else:
        build_app.build(bondline=True)

    for f in expected_files:
        assert f.exists(), f"Expected file {f} not found"
    assert "Mass table per material" in caplog.text


@pytest.mark.skip("seems a double test, output already checked in test_anba")
def test_2d_analysis(temp_workdir, app_state, caplog):
    """Test 2D analysis, mocking ANBA4 subprocess."""
    tmp_path, yml_path = temp_workdir
    caplog.set_level(logging.INFO)

    two_d_app = TwoDApp(app_state, yml_path)

    # Ensure build outputs exist
    build_app = BuildApp(app_state, yml_path)
    prefix = app_state.load_yaml(yml_path)["general"].get("prefix", "b3p")
    workdir = app_state.get_workdir()
    joined_vtu = workdir / "drape" / f"{prefix}_joined.vtu"

    if not joined_vtu.exists():
        build_app.build(bondline=True)

    # Mock subprocess.run to avoid running ANBA4
    with patch("subprocess.run") as mock_run:
        mock_run.return_value.returncode = 0
        two_d_app.mesh2d(rotz=0.0, parallel=False)
        two_d_app.run_anba4(anba_env="anba4-env")

    expected_files = [
        workdir / "2d" / "msec_1000.xdmf",
        workdir / "2d" / "msec_80000.xdmf",
    ]
    for f in expected_files:
        assert f.exists(), f"Expected file {f} not found"
    assert "ANBA4 script completed successfully" in caplog.text


def test_ccx_analysis(temp_workdir, app_state, caplog):
    """Test CCX analysis, mocking subprocess."""
    tmp_path, yml_path = temp_workdir
    caplog.set_level(logging.INFO)

    ccx_app = CcxApp(app_state, yml_path)

    # Ensure build outputs exist
    build_app = BuildApp(app_state, yml_path)
    prefix = app_state.load_yaml(yml_path)["general"].get("prefix", "b3p")
    workdir = app_state.get_workdir()
    joined_vtu = workdir / "drape" / f"{prefix}_joined.vtu"

    if not joined_vtu.exists():
        build_app.build(bondline=True)

    # Mock subprocess.run to avoid running CCX
    with patch("subprocess.run") as mock_run:
        mock_run.return_value.returncode = 0
        ccx_app.ccx(bondline=True)

    expected_files = [
        workdir / "fea" / f"{prefix}_ccx_lc_forward_flap.inp",
    ]
    for f in expected_files:
        assert f.exists(), f"Expected file {f} not found"
    assert f"Written: {workdir}/fea" in caplog.text


def test_full_workflow(temp_workdir, app_state, caplog):
    """Test the full workflow in sequence."""
    tmp_path, yml_path = temp_workdir
    caplog.set_level(logging.INFO)

    # Clean
    clean_app = CleanApp(app_state, yml_path)
    clean_app.clean()

    # Build
    build_app = BuildApp(app_state, yml_path)
    prefix = app_state.load_yaml(yml_path)["general"].get("prefix", "b3p")
    workdir = app_state.get_workdir()
    build_app.build(bondline=True)

    # 2D Analysis
    two_d_app = TwoDApp(app_state, yml_path)
    with patch("subprocess.run") as mock_run:
        mock_run.return_value.returncode = 0
        two_d_app.mesh2d(rotz=0.0, parallel=False)
        two_d_app.run_anba4(anba_env="anba4-env")

    # CCX Analysis
    ccx_app = CcxApp(app_state, yml_path)
    with patch("subprocess.run") as mock_run:
        mock_run.return_value.returncode = 0
        ccx_app.ccx(bondline=False)

    # Verify key outputs
    expected_files = [
        workdir / "drape" / f"{prefix}_joined.vtu",
        workdir / "2d" / "msec_1000.xdmf",
        workdir / "fea" / f"{prefix}_ccx_lc_forward_flap.inp",
    ]
    for f in expected_files:
        assert f.exists(), f"Expected file {f} not found"



import pytest
import os
import shutil
from pathlib import Path
from b3p.cli.app_state import AppState
from b3p.cli.build_app import BuildApp


@pytest.fixture(scope="session")
def temp_example_dir(tmp_path_factory):
    """Fixture to create a temporary copy of the examples directory, excluding temp_* workdirs."""
    base_dir = Path(__file__).parent.parent  # /home/wr1/projects/b3p/
    example_dir = base_dir / "examples"  # /home/wr1/projects/b3p/examples/
    if not example_dir.exists():
        raise FileNotFoundError(f"Examples directory not found at {example_dir}")

    # Create a base temporary directory
    tmp_base = tmp_path_factory.mktemp("build_examples")
    tmp_dir = tmp_base / "examples"

    # Copy examples dir, excluding temp_* directories
    ignore_patterns = shutil.ignore_patterns("temp_*")
    shutil.copytree(example_dir, tmp_dir, ignore=ignore_patterns)

    # Copy test data to tmp_base/data/
    data_dir = base_dir / "tests" / "data"
    if data_dir.exists():
        shutil.copytree(data_dir, tmp_base / "data", dirs_exist_ok=True)
        print(f"Copied test data from {data_dir} to {tmp_base / 'data'}")
    else:
        print(f"Warning: Test data directory {data_dir} not found")

    return tmp_dir


@pytest.fixture(scope="session")
def built_blade(temp_example_dir):
    """Fixture to build the blade once and provide the resulting workdir."""
    original_dir = os.getcwd()
    os.chdir(temp_example_dir)
    try:
        state = AppState()
        yml_path = Path("blade_test.yml")

        build_app = BuildApp(state, yml_path)
        build_app.build()
        workdir = temp_example_dir / "temp_blade"  # New default workdir
        assert workdir.exists(), f"Blade build failed: workdir {workdir} not created"
        yield {
            "workdir": workdir,
            "temp_dir": temp_example_dir.parent,  # Expose parent temp dir for data access
        }
    finally:
        os.chdir(original_dir)



#! /usr/bin/env python3
import pyvista as pv
import argparse
import multiprocessing
import os
import logging

logger = logging.getLogger(__name__)


def conv3d_2d(mesh):
    """Hack to translate a 3D mesh to 2D format, which is required by fenics/anba.

    This function reads a mesh file, modifies its content to convert it from
    3D to 2D format, and writes the modified content back to the file.

    Parameters:
    mesh (str): The path to the mesh file.

    The function performs the following modifications:
    - Removes occurrences of "0.0000000e+00\n".
    - Replaces ' 3"' with ' 2"'.
    - Replaces "XYZ" with "XY"."""
    x = open(mesh, "r").read()
    nb = x[: x.find("Topology")]
    rest = x[x.find("Topology") :]
    nb = nb.replace("0.0000000e+00\n", "\n")
    nb = nb.replace(' 3"', ' 2"')
    nb = nb.replace("XYZ", "XY")
    open(mesh, "w").write(nb + rest)


def vtp2xdmf(vtp):
    """
    Convert a VTP file to XDMF format and translate it to 2D.

    Parameters:
    vtp (str): The file path of the input VTP file. The file must have a .vtp extension.

    Returns:
    str: The file path of the converted XDMF file.

    Notes:
    - The function checks if the XDMF file already exists. If it does, it skips the conversion.
    - The function reads the VTP file, triangulates the mesh, sets the z-coordinates to 0 to make it 2D,
      and saves the mesh in XDMF format.
    - The function calls `conv3d_2d` to further process the XDMF file.
    - Prints messages indicating the status of the conversion.
    """
    assert vtp.endswith(".vtp")
    xd = vtp.replace(".vtp", ".xdmf")
    if os.path.exists(xd):
        logger.info(f"ANBA mesh {xd} already exists - skipping")
        return xd

    mesh = pv.read(vtp)
    tri = mesh.triangulate()  # fenics doesn't do mixed element types
    tri.save(vtp.replace(".vtp", "_tri.vtp"))
    tri.points[:, 2] = 0
    pv.save_meshio(xd, tri, data_format="XML")
    conv3d_2d(xd)
    logger.info(f"converted {vtp} to {xd}")
    return xd


def anba4_prep(section_meshes, parallel=True):
    """
    Convert a list of section meshes to XDMF format.

    This function takes a list of section meshes and converts each mesh to
    XDMF format using multiprocessing for parallel processing.

    Args:
        section_meshes (list): A list of section meshes to be converted.

    Returns:
        list: A list of converted section meshes in XDMF format.
    """
    logger.info("converting section meshes to XDMF")
    if parallel:
        p = multiprocessing.Pool()
        xds = p.map(vtp2xdmf, section_meshes)
        p.close()
        p.join()
    else:
        xds = []
        for i in section_meshes:
            xds.append(vtp2xdmf(i))
    logger.info("done converting section meshes to XDMF")
    return xds


def main():
    """
    Main function to translate section meshes from VTK to XDMF and 2D format.

    This function sets up an argument parser to handle command-line arguments
    for translating section meshes. It expects section meshes in the .vtp format
    and processes them accordingly.

    Command-line Arguments:
    sections (list of str): Section meshes in .vtp format.

    Returns:
    The result of the anba4_prep function, which processes the provided section meshes.
    """
    p = argparse.ArgumentParser(
        description=__doc__
        # "translate section meshes from vtk to XDMF and 2D"
    )
    p.add_argument("sections", nargs="*", help="section meshes in .vtp format")
    args = p.parse_args()
    return anba4_prep(args.sections)


if __name__ == "__main__":
    main()



#! /usr/bin/env python

import numpy as np
import re
import os
import logging
from ruamel.yaml import YAML

logger = logging.getLogger(__name__)


def load_airfoil(af):
    """Load an airfoil from an xfoil formatted text file."""
    if not os.path.exists(af):
        logger.warning(f"Airfoil file {af} not found, keeping as reference")
        return None, None
    with open(af, "r") as f:
        name = f.readline().strip()
        floats = re.findall(r"\d+\.\d+", name)
        if len(floats) == 2:
            offset = 0
            name = af.split("/")[-1].split(".")[0]
        elif name == "":
            offset = 1
            name = af.split("/")[-1].split(".")[0]
        else:
            offset = 1
    return name, np.loadtxt(af, skiprows=offset).tolist()


def load_airfoils(afs, prefix=""):
    """Load a list of airfoils from a dictionary of xfoil formatted text files."""
    dct = {}
    for i in afs:
        if "xy" in afs[i]:
            dct[i] = afs[i]
            continue
        af_path = os.path.join(prefix, afs[i])
        name, xy = load_airfoil(af_path)
        if name is None:
            dct[i] = {"path": afs[i]}
        else:
            dct[i] = {"xy": xy, "name": name, "path": afs[i]}
            logger.info(f"Imported airfoil {name} at thickness {i}")
    return dct


def yaml_make_portable(yaml_file):
    """Import a yaml file and linked files, skipping missing files."""
    logger.info(f"Loading yaml file {yaml_file}")
    prefix = os.path.dirname(yaml_file)

    yaml = YAML()
    d = yaml.load(open(yaml_file, "r"))

    if "aero" in d and "airfoils" in d["aero"]:
        d["aero"]["airfoils"] = load_airfoils(d["aero"]["airfoils"], prefix=prefix)

    subsections = ["materials", "loads", "laminates"]
    for s in subsections:
        if s in d and isinstance(d[s], str) and d[s].endswith(".yml"):
            sub_path = os.path.join(prefix, d[s])
            if os.path.exists(sub_path):
                logger.info(f"Loading {s} from: {sub_path}")
                d[s] = yaml.load(open(sub_path, "r"))
            else:
                logger.warning(f"File {sub_path} not found, keeping as reference")
                d[s] = {"path": d[s]}

    return d


def save_yaml(of, dct):
    """Save dictionary to yaml file."""
    yaml = YAML()
    yaml.default_flow_style = True
    with open(of, "w") as f:
        yaml.dump(dct, f)
        logger.info(f"Written to: {of}")


def save_yaml_portable(yaml_file):
    """Save portable yaml file."""
    d = yaml_make_portable(yaml_file)
    of = os.path.join(os.path.dirname(yaml_file), "portable.yml")
    save_yaml(of, d)



import pytest
import logging
from pathlib import Path
import os
import pandas as pd
from b3p.cli.app_state import AppState
from b3p.cli.build_app import BuildApp
from b3p.cli.ccblade_app import CCBladeApp, has_ccblade

# Configure logging
logging.basicConfig(
    level=logging.INFO, format="%(asctime)s [%(levelname)s] %(message)s"
)
logger = logging.getLogger(__name__)


@pytest.fixture(scope="session")
def run_ccblade(temp_example_dir):
    """Fixture to build the blade and run the CCBlade process."""
    if not has_ccblade:
        pytest.skip("CCBlade functionality is not available")

    original_dir = os.getcwd()
    os.chdir(temp_example_dir)
    try:
        state = AppState()
        yml_path = Path("blade_test.yml")

        logger.info("Building blade with YAML: %s", yml_path)
        build_app = BuildApp(state, yml_path)
        build_app.build()

        logger.info("Running CCBlade analysis")
        ccblade_app = CCBladeApp(state, yml_path)
        ccblade_app.ccblade()

        workdir = temp_example_dir / "temp_blade"
        logger.info("CCBlade workdir: %s", workdir)
        yield {
            "workdir": workdir,
            "temp_dir": temp_example_dir.parent,
        }
    finally:
        os.chdir(original_dir)


def test_ccblade_output(run_ccblade):
    """Test if CCBlade generates the expected output table matching the reference."""
    workdir = run_ccblade["workdir"]
    temp_dir = run_ccblade["temp_dir"]

    generated_file = workdir / "ccblade_output.csv"
    logger.info("Checking for generated CCBlade output: %s", generated_file)
    assert (
        generated_file.exists()
    ), f"CCBlade should generate {generated_file} in the working directory"

    reference_file = temp_dir / "data" / "ccblade_output.csv"
    logger.info("Reference CCBlade output: %s", reference_file)
    assert (
        reference_file.exists()
    ), f"Reference file {reference_file} not found in temp data directory"

    generated_df = pd.read_csv(generated_file)
    reference_df = pd.read_csv(reference_file)

    assert generated_df.equals(
        reference_df
    ), "Generated CCBlade output does not match reference file"



#!/usr/bin/env python3

import yaml
import argparse
import sys


def convert_airfoil_yaml(input_file, output_file):
    """Load YAML and rewrite it using standard PyYAML formatting."""
    # Load input YAML
    try:
        with open(input_file, "r") as f:
            data = yaml.safe_load(f)
    except FileNotFoundError:
        print(f"Error: Input file {input_file} not found")
        sys.exit(1)
    except yaml.YAMLError as e:
        print(f"Error: Invalid YAML in {input_file}: {e}")
        sys.exit(1)

    # Save output YAML with standard formatting
    try:
        with open(output_file, "w") as f:
            yaml.safe_dump(data, f, default_flow_style=None)
        print(f"Successfully wrote converted YAML to {output_file}")
    except Exception as e:
        print(f"Error: Failed to write output file {output_file}: {e}")
        sys.exit(1)


def main():
    # Set up argument parser
    parser = argparse.ArgumentParser(
        description="Rewrite YAML file using standard PyYAML formatting."
    )
    parser.add_argument(
        "-i",
        "--input-yaml",
        type=str,
        default="S98_expanded.yml",
        help="Path to the input YAML file (default: S98_expanded.yml)",
    )
    parser.add_argument(
        "-o",
        "--output-yaml",
        type=str,
        default="converted_airfoil.yml",
        help="Path to the output YAML file (default: converted_airfoil.yml)",
    )
    args = parser.parse_args()

    convert_airfoil_yaml(args.input_yaml, args.output_yaml)


if __name__ == "__main__":
    main()



import pytest
import pyvista as pv
import pandas as pd
import logging

logger = logging.getLogger(__name__)


@pytest.fixture(scope="session")
def load_geometry(built_blade):
    """Fixture to load the joined geometry from the build."""
    workdir = built_blade["workdir"]
    joined_vtu = workdir / "drape" / "test_blade_joined.vtu"
    return pv.read(joined_vtu)


def test_planform(built_blade):
    """Test if the planform CSV matches expected values."""
    workdir = built_blade["workdir"]
    temp_dir = built_blade["temp_dir"]
    planform = workdir / "mesh" / "test_blade_sca_50.csv"
    planform = pd.read_csv(planform, sep=";")

    csv_path = temp_dir / "data" / "test_blade_csv.csv"
    if not csv_path.exists():
        pytest.skip("Reference CSV 'test_blade_csv.csv' not found in tests/data/")
    expected_df = pd.read_csv(csv_path, sep=";")

    logger.info(f"Expected DataFrame:\n{expected_df}")
    pd.testing.assert_frame_equal(expected_df, planform)


def test_geometry_bounding_box(load_geometry):
    """Test if the geometry bounding box is correct."""
    assert load_geometry.bounds == pytest.approx(
        (
            -3.0369677543640137,
            2.499872922897339,
            -3.3991336822509766,
            3.0669267177581787,
            0.0,
            100.0,
        ),
        rel=1e-5,
    ), "Bounding box does not match expected values"



#!/usr/bin/env python3
import argparse
import json
import logging
import multiprocessing
import os
import traceback
import numpy as np
import pyvista as pv
from dolfin import Mesh, XDMFFile, MeshFunction
from anba4 import (
    material,
    anbax,
    DecoupleStiffness,
    PrincipalAxesRotationAngle,
    ComputeMassCenter,
    ComputeTensionCenter,
    ComputeShearCenter,
)

# Configure logging
logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s")
logger = logging.getLogger(__name__)


def get_material_db(material_map, unit_factor=1):
    """Load and process a material database from a JSON material map file."""
    if not os.path.isfile(material_map):
        raise FileNotFoundError(f"Material map file not found: {material_map}")

    with open(material_map, "r") as f:
        mm1 = json.load(f)
        mm = mm1.get("map", mm1)  # Support both old and new format
        mat_db = mm1.get("matdb", None)

    if "-1" in mat_db:
        mm["-1"] = -1

    materials = {}
    for mat_name, mesh_id in mm.items():
        if mat_name == "matdb":
            continue
        matdb_id = mat_name
        if matdb_id not in mat_db:
            logger.error(
                f"Material ID {matdb_id} from material_map not found in materials.yaml"
            )
            raise KeyError(f"Material ID {matdb_id} not found in materials.yaml")
        matdb_entry = mat_db[matdb_id]
        logger.debug(
            f"Processing material {matdb_id} (mesh ID {mesh_id}): {matdb_entry}"
        )
        density = matdb_entry.get("rho", matdb_entry.get("density", 1.0))

        if "e11" in matdb_entry:
            required_keys = [
                "e11",
                "e22",
                "e33",
                "g12",
                "g13",
                "g23",
                "nu12",
                "nu13",
                "nu23",
            ]
            if not all(k in matdb_entry for k in required_keys):
                raise ValueError(
                    f"Missing orthotropic properties for material {matdb_id}: {required_keys}"
                )
            matMechanicProp = np.zeros((3, 3))
            matMechanicProp[0, 2] = matdb_entry["e11"] * unit_factor
            matMechanicProp[0, 1] = matdb_entry["e22"] * unit_factor
            matMechanicProp[0, 0] = matdb_entry["e33"] * unit_factor
            matMechanicProp[1, 2] = matdb_entry["g23"] * unit_factor
            matMechanicProp[1, 1] = matdb_entry["g13"] * unit_factor
            matMechanicProp[1, 0] = matdb_entry["g12"] * unit_factor
            matMechanicProp[2, 2] = matdb_entry["nu23"]
            matMechanicProp[2, 1] = matdb_entry["nu13"]
            matMechanicProp[2, 0] = matdb_entry["nu12"]
            materials[matdb_id] = material.OrthotropicMaterial(matMechanicProp, density)

        else:
            if "E" not in matdb_entry and "Ex" not in matdb_entry:
                raise ValueError(
                    f"Missing 'E' or 'Ex' for isotropic material {matdb_id}. Entry: {matdb_entry}"
                )
            if "nu" not in matdb_entry:
                raise ValueError(
                    f"Missing 'nu' for isotropic material {matdb_id}. Entry: {matdb_entry}"
                )
            E = matdb_entry.get("E", matdb_entry.get("Ex")) * unit_factor
            nu = min(matdb_entry["nu"], 0.49)
            materials[matdb_id] = material.IsotropicMaterial([E, nu], density)

    mm_inv = {str(v): k for k, v in mm.items() if k != "matdb"}
    logger.debug(f"Material map inverse: {mm_inv}")

    return materials, mm_inv


def export_unit_strains(anba, result_file_name, pv_mesh):
    """Export strain fields for unit displacements to separate VTK files."""

    for disp, suffix in [([1, 0, 0], "mx"), ([0, 1, 0], "my"), ([0, 0, 1], "mz")]:
        anba.strain_field([0, 0, 0], disp, "local", "paraview")
        nodevalues = anba.STRAIN.vector().get_local().reshape(-1, 6)
        pv_mesh.cell_data[f"strain_{suffix}"] = nodevalues

    pv_mesh.save(result_file_name)


def solve_anba4(mesh_filename, material_db_filename):
    """Solve the ANBA4 problem for a mesh and material database."""
    if not mesh_filename.endswith(".xdmf"):
        logger.error(f"Expected .xdmf mesh file, got {mesh_filename}")
        return
    if not os.path.isfile(mesh_filename):
        logger.error(f"Mesh file not found: {mesh_filename}")
        return
    if not os.path.isfile(material_db_filename):
        logger.error(f"Material database file not found: {material_db_filename}")
        return

    json_output_filename = f"{mesh_filename}.json"
    if os.path.isfile(json_output_filename):
        logger.info(f"Output file {json_output_filename} exists, skipping")
        return

    logger.info(f"Processing {mesh_filename}")
    material_db, mm_inv = get_material_db(material_db_filename)

    with XDMFFile(mesh_filename) as infile:
        mesh = Mesh()
        infile.read(mesh)
    # Read .xdmf directly with PyVista, handling potential lack of time steps
    pv_mesh = pv.read(mesh_filename.replace(".xdmf", "_tri.vtp"))

    materials = MeshFunction("size_t", mesh, mesh.topology().dim())
    fiber_orientations = MeshFunction("double", mesh, mesh.topology().dim())
    plane_orientations = MeshFunction("double", mesh, mesh.topology().dim())

    material_unique_indices = np.unique(pv_mesh.cell_data["mat"])
    material_unique_indices.sort()
    logger.debug(f"Mesh material IDs: {material_unique_indices.tolist()}")

    material_map = {}
    for mesh_id in material_unique_indices:
        mesh_id_str = str(mesh_id)
        if mesh_id_str not in mm_inv:
            logger.error(
                f"Mesh material ID {mesh_id_str} not found in material map inverse: {mm_inv}"
            )
            raise KeyError(f"Mesh material ID {mesh_id_str} not found in material map")
        material_map[mesh_id] = mm_inv[mesh_id_str]
    logger.debug(f"Material mapping: {material_map}")

    anba_map = {mesh_id: idx for idx, mesh_id in enumerate(material_unique_indices)}
    logger.debug(f"ANBA4 mapping (mesh ID to index): {anba_map}")

    material_ids = [anba_map[i] for i in pv_mesh.cell_data["mat"].tolist()]
    logger.debug(f"ANBA4 material IDs: {np.unique(material_ids).tolist()}")

    material_library = [
        material_db[material_map[mesh_id]] for mesh_id in material_unique_indices
    ]
    logger.debug(
        f"Material library order: {[material_map[mesh_id] for mesh_id in material_unique_indices]}"
    )

    materials.set_values(material_ids)
    fiber_orientations.set_all(0.0)  # TODO: Support off-axis laminates
    plane_orientations.set_values(pv_mesh.cell_data["angle2"].tolist())

    anba = anbax(
        mesh, 1, material_library, materials, plane_orientations, fiber_orientations
    )
    stiffness = anba.compute()
    stiffness_matrix = stiffness.getDenseArray()

    mass = anba.inertia()
    mass_matrix = mass.getDenseArray()

    decoupled_stiffness = DecoupleStiffness(stiffness)
    principal_axes_rotation_angle = PrincipalAxesRotationAngle(decoupled_stiffness)
    mass_center = ComputeMassCenter(mass)
    tension_center = ComputeTensionCenter(stiffness_matrix)
    shear_center = ComputeShearCenter(stiffness_matrix)

    output = {
        "name": mesh_filename,
        "stiffness": stiffness_matrix.tolist(),
        "mass_matrix": mass_matrix.tolist(),
        "decoupled_stiffness": decoupled_stiffness.tolist(),
        "principal_axes_rotation": principal_axes_rotation_angle,
        "mass_center": mass_center,
        "tension_center": tension_center,
        "shear_center": shear_center,
    }
    with open(json_output_filename, "w") as f:
        json.dump(output, f, indent=4)

    result_file = mesh_filename.replace(".xdmf", "_results.vtp")
    export_unit_strains(anba, result_file, pv_mesh)
    # import pyvista as pv
    # mesh = pv.read(result_filename)
    # mesh.save(result_filename.replace(".xdmf", ".vtu"))
    logger.info(f"Wrote results to {json_output_filename} and {result_file}")


def run_solver(mesh, material_db):
    """Wrapper to run solve_anba4 with error handling for multiprocessing."""
    try:
        solve_anba4(mesh, material_db)
    except Exception as e:
        logger.error(f"Failed to process {mesh}: {str(e)}\n{traceback.format_exc()}")


def main(args):
    """Process multiple meshes using ANBA4 solver in parallel."""
    with multiprocessing.Pool(processes=multiprocessing.cpu_count()) as pool:
        pool.starmap(run_solver, [(mesh, args.material_db) for mesh in args.meshes])


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Run ANBA4 solver on a series of mesh files."
    )
    parser.add_argument("meshes", nargs="*", help="List of XDMF mesh files")
    parser.add_argument("material_db", help="Material map JSON file")
    args = parser.parse_args()
    main(args)



import pyvista as pv


def test_bondline_model(built_blade):
    """Test if the bondline model is created."""
    workdir = built_blade["workdir"]
    bondline_vtu = list(workdir.glob("drape/*bondline.vtu"))
    assert bondline_vtu, "Bondline VTU file should exist"
    vtu = pv.read(bondline_vtu[0])
    assert vtu.n_points > 0, "Bondline VTU should contain points"



import logging
from pathlib import Path

logger = logging.getLogger(__name__)

try:
    from b3p.bem import ccblade_run

    has_ccblade = True
except ImportError:
    logger.warning("Could not import ccblade_run. Functionality will be disabled.")
    has_ccblade = False


class CCBladeApp:
    def __init__(self, state, yml: Path):
        self.state = state
        self.yml = yml

    def ccblade(self):
        if has_ccblade:
            ccblade = ccblade_run.ccblade_run(self.yml)
            ccblade.run()
        else:
            logger.error("ccblade_run is not available.")



import json
from pathlib import Path

import pytest
import numpy as np
from b3p.cli.app_state import AppState
from b3p.cli.two_d_app import TwoDApp
import meshio


@pytest.fixture
def two_d_app(built_blade):
    """Provide a fresh TwoDApp instance for each test."""
    state = AppState.get_instance()
    yml_path = built_blade["temp_dir"] / "examples" / "blade_test.yml"
    return TwoDApp(state, yml_path)


def test_mesh2d_output(two_d_app):
    """Test that mesh2d generates a non-None result for test.yml after blade build."""
    result = two_d_app.mesh2d(parallel=False)  # Generate meshes
    print(f"mesh2d result: {result}")
    assert result is not None, "mesh2d should return a non-None result"


def test_anba4_output_json(two_d_app):
    """Test that ANBA4 produces at least one JSON output file with valid content after blade build."""
    try:
        two_d_app.mesh2d(parallel=False)  # Generate meshes
        two_d_app.run_anba4()  # Run ANBA4 analysis

        prefix = two_d_app.state.get_prefix("drape")

        output_jsons = list(prefix.parent.glob("2d/msec_*.json"))
        assert output_jsons, "ANBA4 should create at least one output JSON file"

        with open(output_jsons[0], "r") as f:
            json_data = json.load(f)
        print(f"Found JSON files: {output_jsons}")
        print(f"First JSON content: {json_data}")
    finally:
        pass


def test_anba4_output_file(two_d_app):
    """Test that ANBA4 creates the expected XDMF output file after blade build."""
    try:
        two_d_app.mesh2d(parallel=False)  # Generate meshes
        two_d_app.run_anba4()  # Run ANBA4 analysis
        prefix = two_d_app.state.get_prefix("drape")

        output_file = Path(prefix).parent / "2d" / "msec_100000.xdmf"
        assert output_file.exists(), "ANBA4 should create msec_100000.xdmf"
    finally:
        pass


def test_mesh2d_compare_reference(two_d_app, built_blade):
    """Test that 2D meshes match reference meshes in tests/data/reference_2d_meshes/ after blade build."""
    try:
        # Run mesh2d to generate meshes
        two_d_app.mesh2d(parallel=False)

        prefix = two_d_app.state.get_prefix("drape")

        print(f"Using prefix: {prefix}")
        # Get generated mesh files
        generated_meshes = list(Path(prefix.parent).glob("2d/msec_*0.xdmf"))

        print(f"Generated meshes: {generated_meshes}")
        assert generated_meshes, "mesh2d should generate at least one .xdmf file"

        # Reference directory in temporary test data
        ref_dir = built_blade["temp_dir"] / "data" / "reference_2d_meshes"
        assert ref_dir.exists(), f"Reference directory {ref_dir} not found"

        # Compare each generated mesh with its reference
        for gen_mesh in generated_meshes:
            ref_mesh_path = ref_dir / gen_mesh.name
            assert ref_mesh_path.exists(), f"Reference mesh {ref_mesh_path} not found"

            # Load meshes with pyvista
            gen_mesh_data = meshio.read(gen_mesh)
            ref_mesh_data = meshio.read(ref_mesh_path)

            # Compare mesh bounding boxes
            gen_bounds = gen_mesh_data.points.min()
            ref_bounds = ref_mesh_data.points.min()
            assert np.allclose(
                gen_bounds, ref_bounds, rtol=1e-1, atol=1e-2
            ), f"Bounding boxes in {gen_mesh.name} do not match reference"

            # Compare number of points
            assert len(gen_mesh_data.points) == len(
                ref_mesh_data.points
            ), f"Number of points in {gen_mesh.name} does not match reference"
    finally:
        pass



#! /usr/bin/env python3

import numpy as np

# import yaml
from ruamel.yaml import YAML
import os
from itertools import chain, zip_longest
import pickle
import json
import copy as cp
from sympy import symbols, sympify
import logging
from pathlib import Path

logger = logging.getLogger(__name__)


def plyify(r, t, ply_thickness, reverse=False):
    """Fill thickness distribution using plies.

    params:
        r (list): radius distribution
        t (list): thickness distribution
        ply_thickness (float): thickness of the plies
        reverse (bool): flag determining whether the plies are numbered from the
            top or bottom of the stack
        returns:    list of plies in the form [rmin, rmax]"""
    active = []
    done = []
    np = len(r)
    for i in range(np):
        while t[i] > len(active) * ply_thickness:
            active.append([r[i], -1])
        while t[i] <= (len(active) - 1) * ply_thickness:
            pop = active.pop()
            pop[-1] = r[i]
            done.append(pop)
    for j in active:
        j[-1] = r[i]
        done.append(j)
    for i in done:
        i[0] = round(i[0] * 1e3, -1) * 1e-3
        i[1] = round(i[1] * 1e3, -1) * 1e-3
    # reverse the stack so that longest plies get the lowest ply numbers
    return done if reverse else list(reversed(done))


def expand_chamfered_core(core):
    rr, tr = np.array(core["slab"]).T

    coordinates = []
    added_cores = []
    for chmf in core["chamfers"]:
        ns = chmf["nstep"] if "nstep" in chmf else 4

        idx = chmf["id"]

        fr = np.array(chmf["ratio"])

        intrr = np.sort(np.unique(np.hstack([rr, fr[:, 0]])))

        thickness = np.interp(intrr, rr, tr)

        ratio = np.interp(intrr, fr[:, 0], fr[:, 1])

        distance = 1e-3 * ratio * thickness

        # chamfer_step = "chstep"
        added_coordinates = {}
        for i in range(0, ns + 1):
            key = f"{idx[0]}_c{i}"
            added_coordinates[key] = {
                "base": idx[0],
                "points": np.stack(
                    [intrr, (-1.0 if idx[1] == 0 else 1.0) * i * distance / ns]
                ).T.tolist(),
            }

        coordinates.append(added_coordinates)

        for j in range(0, ns):
            thic = thickness * ((j + 1.0) / (ns + 1.0))
            nc = cp.deepcopy(core)
            del nc["chamfers"]

            nc["slab"] = np.stack([intrr, thic]).T.tolist()

            bc = nc["cover"][f"{idx[0]}"]
            if idx[1] == 0:
                nc["cover"][f"{idx[0]}_c{j}"] = bc
                nc["cover"][f"{idx[0]}_c{j + 1}"] = [-1, bc[0], 0]

            # chamfers from 2nd coordinate
            elif idx[1] == 1:
                nc["cover"][f"{idx[0]}_c{j}"] = bc
                nc["cover"][f"{idx[0]}_c{j + 1}"] = [bc[1], -bc[0], 0]

            del nc["cover"][f"{idx[0]}"]

            added_cores.append(nc)

        core["cover"][key] = cp.deepcopy(core["cover"][idx[0]])

        if idx[0] in core["cover"]:
            del core["cover"][idx[0]]

    del core["chamfers"]
    return coordinates, added_cores


def expand_chamfered_cores(bldict, outputfile: Path = None):
    """ "Expand chamfered cores into laminate strips and write out to a new YAML file."""
    dctcopy = cp.deepcopy(bldict)
    for i in bldict["laminates"]["slabs"]:
        slab = dctcopy["laminates"]["slabs"][i]
        if "chamfers" in slab:
            coordinates, cores = expand_chamfered_core(slab)
            for j in range(len(cores)):
                slabname = i + f"_ch{j}"
                dctcopy["laminates"]["slabs"][slabname] = cores[j]

            for j in range(len(coordinates)):
                for k in coordinates[j]:
                    dctcopy["mesh"]["coordinates"][k] = coordinates[j][k]

    yaml = YAML()
    yaml.dump(dctcopy, outputfile.open("w"))
    if outputfile:
        logger.info(f"written expanded chamfered cores to {outputfile}")
    else:
        logger.warning(
            "No output file specified, not writing expanded chamfered cores."
        )
    return dctcopy


def ply_stack(r, t, t_ply=1.0, reverse=False, subdivisions=5000, material=11):
    """
    Taking in a r,t thickness over radius distribution, split it up in
    plies, note that it rounds down in thickness (if the last ply does not fit
    in the thickness it does not go on), so if you have thick plies,
    make sure that the mm thickness is slightly up from a whole number of plies.

    :param r: radius distribution
    :param t: thickness distribution
    :param t_ply: thickness of the plies
    :param reverse: flag determining whether the plies are numbered from the
        top or bottom of the stack
    :param subdivisions: number of subdivisions
    :param material: material number
    """
    x = np.linspace(min(r), max(r), subdivisions)
    y = np.interp(x, list(r), t)
    st = plyify(x, y, t_ply, reverse)
    return [[material, t_ply] + i for i in st]


def coreblock(r, t, subdivisions=200, material=11):
    """Make a thickness distribution into a block divisions.

    :param r: radius distribution
    :param t: thickness distribution
    :param subdivisions: number of subdivisions
    :param material: material number"""
    assert len(r) == len(t)
    lr = len(r)
    x = np.array(sorted(list(np.linspace(min(r), max(r), int(subdivisions))) + list(r)))
    np.interp(0.5 * (x[:-1] + x[1:]), list(r), t)

    stack = []
    for i in range(lr - 1):
        rmin, rmax = r[i], r[i + 1]
        tmin, tmax = t[i], t[i + 1]
        tt = tmin if tmin == tmax else 0.5 * (tmin + tmax)
        if tt > 0:
            stack.append([material, tt, rmin, rmax])

    return stack


def number_stack(stack, splitstack, key, increment):
    """Create ply numbering for the plies in the stack, depends on the start key
    and increment as well as the above/below ratio (splitstack).

    :param stack: list of plies
    :param splitstack: ratio of above/below plies
    :param key: start key from top and bottom
    :param increment: increment, number by which to increment the key (can be used to interleave plies)
    """
    sp = np.nan_to_num(splitstack.astype(float))
    if sp.sum() != 1.0:
        exit(f"sum {splitstack} not 1.")

    ky = np.nan_to_num(key.astype(float)).astype(int)
    inc = np.nan_to_num(increment.astype(float), nan=1).astype(int)

    ss = (len(stack) * sp).round().astype(int)
    st, sb = np.array(range(ss[0])), np.array(range(ss[1]))
    stt = ky[0] + st * inc[0]
    sbt = ky[1] + sb * inc[1]
    return [i for i in chain.from_iterable(zip_longest(stt, sbt)) if i is not None]


from sympy import lambdify


def get_coverage(slab, datums, rr):
    """
    Get the coverage of the slab and insert the datums.
    :param slab: slab dictionary
    :param datums: datums dictionary
    :param rr: radius distribution
    """
    assert "cover" in slab
    cov = slab["cover"]

    # Create symbolic variables for all datum keys
    symbols_dict = {k: symbols(k) for k in datums}

    for d in cov:
        cc = cov[d]
        # Parse the expressions symbolically
        idd = [sympify(cc[0], locals=symbols_dict), sympify(cc[1], locals=symbols_dict)]

        for j in range(2):
            expr = idd[j]  # Symbolic expression

            # Replace symbols with interpolated numerical arrays
            substitutions = {}
            for key, sym in symbols_dict.items():
                if sym in expr.free_symbols:  # Check if the symbol is present
                    xy = np.array(datums[key]["xy"])
                    dst = np.interp(
                        rr,
                        xy[:, 0] / datums[key]["scalex"],
                        xy[:, 1] * datums[key]["scaley"],
                    )
                    substitutions[sym] = dst  # Store the substitution separately

            # Perform the substitutions numerically after creating a callable function
            func = lambdify(list(substitutions.keys()), expr, modules="numpy")
            idd[j] = func(*substitutions.values())  # Pass the numerical arrays

        cov[d][0], cov[d][1] = idd  # Assign the evaluated results

    return cov


def add_bondline_material(matdb, material_map):
    """
    Add a bondline material to the material map, if it is not already there,
    this is needed because the 2d mesher adds elements with -1 material ID to connect the webs to the shell

    params:
        matdb (dict): material database
        material_map (dict): material map
        returns:
            material_map (dict): material map with bondline material added
    """
    possible_glue_names = ["adhesive", "bonding"]
    for i in matdb:
        for j in possible_glue_names:
            if matdb[i]["name"].find(j) != -1:
                matkey = i
                material_map["bondline"] = -1

    if "bondline" in material_map:
        matdb["bondline"] = cp.deepcopy(matdb[matkey])
        matdb["bondline"]["name"] = "bondline"

    return material_map


def export_matdb(blade, material_map):
    """Export material database to JSON."""
    workdir = blade["general"]["workdir"]
    matmap = os.path.join(workdir, "drape", "material_map.json")

    if "bondline" in blade["mesh"] and "material" in blade["mesh"]["bondline"]:
        maxid = max(v for v in material_map.values() if isinstance(v, int))
        material_map[blade["mesh"]["bondline"]["material"]] = maxid + 1

    material_map = add_bondline_material(blade["materials"], material_map)

    with open(matmap, "w") as f:
        json.dump({"matdb": blade["materials"], "map": material_map}, f)

    logger.info(f"written material map to {matmap}")


def export_plybook(stacks, outputfile):
    pickle.dump(stacks, open(outputfile, "wb"))
    logger.info(f"written to {outputfile}")


def lamplan2plies(blade, outputfile="__plybook.pck"):
    """Convert a lamplan to a list of plies."""
    root_radius = blade["planform"]["z"][0][1]

    tip_radius = blade["planform"]["z"][-1][1]

    slabs = blade["laminates"]["slabs"]

    datums = blade["laminates"]["datums"] if "datums" in blade["laminates"] else {}

    allstacks = []

    # use a multiple of lamplan length as radius grid to interpolate geometric variables to
    n_s = round(tip_radius * 4)

    radius = np.linspace(0, tip_radius - root_radius, n_s)
    radius_relative = np.linspace(0, 1, n_s)
    material_map = {}

    for i in slabs:
        material = slabs[i]["material"]
        if material not in material_map:
            material_map[material] = len(material_map) + 1

        grid = slabs[i]["grid"]

        cover = get_coverage(slabs[i], datums, radius_relative)

        draping = "plies" if "draping" not in slabs[i] else slabs[i]["draping"]

        ply_thickness = float(
            1.0 if "ply_thickness" not in slabs[i] else slabs[i]["ply_thickness"]
        )

        r, t = np.array(slabs[i]["slab"]).T

        r *= (
            tip_radius - root_radius if "rscale" not in slabs[i] else slabs[i]["rscale"]
        )
        t *= ply_thickness

        if draping == "blocks":
            stack = coreblock(r, t, material=material_map[material])
        elif draping == "plies":
            stack = ply_stack(r, t, ply_thickness, material=material_map[material])

        # assigns keys to the plies in the stack depending on how the stack is split up
        # what increment the ply keys are stacked with (non-1 increment allowing interleaving of plies)
        stack_numbering = number_stack(
            stack,
            (
                np.array([1, 0])
                if "splitstack" not in slabs[i]
                else np.array(slabs[i]["splitstack"])
            ),
            (
                np.array([0, 4000])
                if "key" not in slabs[i]
                else np.array(slabs[i]["key"])
            ),
            (
                np.array([1, -1])
                if "increment" not in slabs[i]
                else np.array(slabs[i]["increment"])
            ),
        )
        allstacks.append(
            {
                "name": i.strip(),
                "grid": grid.strip(),
                "cover": cover,
                "numbering": stack_numbering,
                "stack": stack,
                "r": radius,
            }
        )
    export_matdb(blade, material_map)
    export_plybook(allstacks, outputfile=outputfile)
    return allstacks


def slab2plybook(yamlfile, outputfile="__lamplan.pck"):
    blade = yaml.safe_load(open(yamlfile, "r"))
    lamplan2plies(blade, outputfile)
    logger.info(f"written plydrape to {outputfile}")



import logging
from pathlib import Path
import os
import subprocess
import shutil
from b3p.anba import anba4_prep
from b3p.anba import mesh_2d
import glob

logger = logging.getLogger(__name__)


class TwoDApp:
    def __init__(self, state, yml: Path):
        """Initialize TwoDApp with state and YAML config file."""
        self.state = state
        self.yml = yml

    def mesh2d(self, rotz=0.0, parallel=True):
        """Create 2D meshes from blade sections."""
        dct = self.state.load_yaml(self.yml)
        if "mesh2d" not in dct:
            logger.error("No mesh2d section in yml file")
            return
        if "sections" not in dct["mesh2d"]:
            logger.error("No sections in mesh2d section in yml file")
            return
        sections = dct["mesh2d"]["sections"]

        drape_prefix = self.state.get_prefix("drape")
        mesh_prefix = self.state.get_prefix("mesh")

        section_meshes = mesh_2d.cut_blade_parallel(
            f"{drape_prefix}_joined.vtu",
            sections,
            if_bondline=False,
            rotz=rotz,
            var=f"{mesh_prefix}_variables.json",
            parallel=parallel,
        )
        if not section_meshes or not all(
            Path(mesh).exists() for mesh in section_meshes
        ):
            logger.error("Section meshes were not generated correctly")
            return []
        return anba4_prep.anba4_prep(section_meshes, parallel=parallel)

    def run_anba4(self, anba_env="anba4-env"):
        """Run ANBA4 on 2D meshes."""
        yml = self.yml
        self.state.load_yaml(yml)
        prefix = self.state.get_prefix("drape")
        meshes = glob.glob(str(Path(prefix) / "2d" / "msec_*.xdmf"))
        conda_path = os.environ.get("CONDA_EXE") or shutil.which("conda")
        if conda_path is None:
            logger.error("Conda not found - please install conda")
            return
        result = subprocess.run(
            [conda_path, "env", "list"], capture_output=True, text=True
        )
        if result.returncode != 0 or anba_env not in result.stdout:
            logger.error(f"Conda environment {anba_env} not found - please create it")
            return

        logger.info(f"Using Conda environment for running anba4 {anba_env}")
        self.state.load_yaml(yml)
        if not meshes:
            meshes = self.mesh2d()

        material_map = str(Path(prefix).parent / "material_map.json")
        script_path = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "..", "anba", "anba4_solve.py")
        )
        logger.info(f"Running ANBA4 using {script_path} in env {anba_env}")
        conda_command = [
            conda_path,
            "run",
            "-n",
            anba_env,
            "python",
            script_path,
            *meshes,
            material_map,
        ]

        logger.info(" ".join(conda_command))
        result = subprocess.run(
            conda_command,
            capture_output=True,
            text=True,
            env={
                **os.environ.copy(),
                "OPENBLAS_NUM_THREADS": "1",
                "MKL_NUM_THREADS": "1",
                "OMP_NUM_THREADS": "1",
                "CUDA_VISIBLE_DEVICES": "-1",
            },
        )
        if result.returncode != 0:
            logger.error(f"ANBA4 script failed with return code {result.returncode}")
            logger.error(f"Stdout: {result.stdout}")
            logger.error(f"Stderr: {result.stderr}")
        else:
            logger.info("ANBA4 script completed successfully")
            logger.debug(f"Stdout: {result.stdout}")
        return result.returncode

    def clean(self):
        """Remove 2D working directory and its contents."""
        self.state.load_yaml(self.yml)
        workdir = Path(self.state.get_prefix("2d")).parent / "2d"
        if not workdir.exists():
            logger.info(f"Workdir {workdir} does not exist - nothing to clean")
            return

        try:
            shutil.rmtree(workdir)
            logger.info(f"Removed workdir {workdir}")
        except Exception as e:
            logger.error(f"Failed to remove workdir {workdir}: {e}")



from pathlib import Path
import os
from b3p.cli import yml_portable
from b3p.laminates import build_plybook
import logging
from typing import Optional, Dict

logger = logging.getLogger(__name__)


class AppState:
    """Singleton to manage application state."""

    _state: Optional["AppState"] = None

    def __init__(self):
        self.dct: Optional[Dict] = None
        self.yml_dir: Optional[Path] = None
        self.workdir_path: Optional[Path] = None  # Stores the resolved workdir as Path

    @classmethod
    def get_instance(cls):
        if cls._state is None:
            cls._state = cls()
        return cls._state

    def load_yaml(self, yml: Path) -> Optional[Dict]:
        if self.dct is None:
            yml = yml.resolve()
            self.yml_dir = yml.parent
            self.dct = yml_portable.yaml_make_portable(yml)
            self._set_workdir()
            self.make_workdir()
            self.expand_chamfered_cores()

        logger.info(f"Loaded YAML data from {self.dct['general']['workdir']}")
        return self.dct

    def _set_workdir(self):
        if self.dct and "general" in self.dct and "workdir" in self.dct["general"]:
            workdir_str = self.dct["general"]["workdir"]
            workdir_path = Path(workdir_str)
            if not workdir_path.is_absolute():
                workdir_path = self.yml_dir / workdir_path
                workdir_path = workdir_path.resolve()
            self.workdir_path = workdir_path
            logger.info(f"Setting workdir to {self.workdir_path}")
        else:
            default_workdir = self.yml_dir / "output"
            self.workdir_path = default_workdir.resolve()

    def expand_chamfered_cores(self):
        if self.dct:
            self.dct = build_plybook.expand_chamfered_cores(
                self.dct,
                self.workdir_path / (self.dct["general"]["prefix"] + "_expanded.yml"),
            )

    def make_workdir(self):
        if self.dct is None:
            raise ValueError("No YAML data loaded")
        if self.workdir_path is None:
            raise ValueError("Workdir not set")
        if not self.workdir_path.is_dir():
            os.makedirs(self.workdir_path, exist_ok=True)

    def get_prefix(self, subdir: Optional[str] = None) -> Optional[Path]:
        if self.dct is None:
            logger.warning("No YAML data loaded, returning None")
            return None
        if self.workdir_path is None:
            logger.warning("Workdir not set, returning None")
            return None
        prefix_name = self.dct["general"].get("prefix", "b3p")
        if subdir is None:
            prefix = self.workdir_path / prefix_name
        else:
            prefix = self.workdir_path / subdir / prefix_name
            if not (self.workdir_path / subdir).is_dir():
                os.makedirs(self.workdir_path / subdir, exist_ok=True)
        return prefix.resolve()

    def get_workdir(self) -> Optional[Path]:
        if self.workdir_path is None:
            logger.warning("Workdir not set, returning None")
            return None
        return self.workdir_path

    def reset(self):
        self.dct = None
        self.yml_dir = None
        self.workdir_path = None



import logging
from pathlib import Path
import os
import pickle
from . import yml_portable
from b3p.geometry import build_blade_geometry
from b3p.mesh import (
    add_load_to_mesh,
    add_te_solids,
    build_blade_structure,
    combine_meshes,
)
from b3p.laminates import build_plybook, drape_mesh, drape_summary
from rich.logging import RichHandler  # Add rich log formatting


logging.basicConfig(handlers=[RichHandler(rich_tracebacks=True)], level=logging.INFO)  # Configure rich for logging
logger = logging.getLogger(__name__)


class BuildApp:
    def __init__(self, state, yml: Path):
        self.state = state
        self.dct = self.state.load_yaml(yml)

    def geometry(self):
        prefix = self.state.get_prefix("mesh")
        build_blade_geometry.build_blade_geometry(self.dct, prefix)
        prefix = self.state.get_prefix()
        yml_portable.save_yaml(f"{prefix}_portable.yml", self.dct)

    def mesh(self):
        prefix = self.state.get_prefix("mesh")
        build_blade_structure.build_blade_structure(self.dct, prefix)

    def drape(self, bondline: bool = True):
        plybookname = "_plybook.pck"
        prefix = self.state.get_prefix("drape")
        mesh_prefix = self.state.get_prefix("mesh")
        pbookpath = str(prefix) + plybookname

        build_plybook.lamplan2plies(self.dct, pbookpath)
        slb = self.dct["laminates"]["slabs"]
        used_grids = {slb[i]["grid"] for i in slb}

        if os.path.exists(pbookpath):
            plybook = pickle.load(open(pbookpath, "rb"))
            meshes = []
            for grid in used_grids:
                out = f"{prefix}_{grid}_dr.vtu"
                logger.info(f"Draping mesh for grid {grid} to {out}")
                drape_mesh.drape_mesh(f"{mesh_prefix}_{grid}.vtp", plybook, grid, out)
                meshes.append(out)
            combine_meshes.combine_meshes(meshes, f"{prefix}_joined.vtu")
            if bondline:
                add_te_solids.add_bondline(self.dct, prefix)
                logger.info("Bondline added to mesh")
        else:
            logger.error(f"Plybook not found at {pbookpath}")
            raise FileNotFoundError(f"Plybook not found at {pbookpath}")

    def mass(self):
        wd = self.state.get_workdir()
        prefix = self.state.get_prefix("drape")
        if not wd.is_dir():
            logger.warning("No workdir found, building new")
            self.build(bondline=True)

        mass_table = drape_summary.drape_summary(f"{prefix}_joined.vtu")
        mass_table.to_csv(f"{prefix}_mass.csv")
        mass_table.replace(to_replace="_", value="", regex=True).to_latex(
            f"{prefix}_mass.tex", index=False
        )
        logger.info("Mass table per material:\n%s", mass_table)

    def apply_loads(self):
        prefix = self.state.get_prefix("drape")
        add_load_to_mesh.add_load_to_mesh(
            self.dct,
            f"{prefix}_joined.vtu",
            f"{prefix}_loads.png",
        )

    def build(self, bondline: bool = True):
        self.geometry()
        self.mesh()
        self.drape(bondline=bondline)
        self.mass()
        self.apply_loads()



import glob
import os
from pathlib import Path
from b3p.cli.app_state import AppState
from b3p.cli.build_app import BuildApp


def test_build_success(built_blade):
    """Test if the build completes successfully."""
    assert built_blade["workdir"].exists(), "Working directory was not created."


def test_build_output_files(built_blade):
    """Test if all expected output files are generated."""
    workdir = built_blade["workdir"]
    output_files = {
        os.path.basename(f) for f in glob.glob(str(workdir / "**/*"), recursive=True)
    }
    expected_files = {
        "test_blade_base.vtp",
        "test_blade.png",
        "test_blade_shell.vtp",
        "test_blade_w0.vtp",
        "test_blade_w1.vtp",
        "test_blade_w2.vtp",
        "test_blade_w3.vtp",
        "test_blade_w4.vtp",
        "test_blade_w3_dr.vtu",
        "test_blade_shell_dr.vtu",
        "test_blade_w4_dr.vtu",
        "test_blade_joined.vtu",
        "material_map.json",
        "test_blade_loads.png",
        "test_blade.vtp",
        "test_blade_mass.csv",
        "test_blade_mass.tex",
    }

    missing_files = expected_files - output_files
    assert not missing_files, f"Missing expected files: {missing_files}"


def test_build_blademass_match(built_blade):
    """Test if the mass table matches the reference."""
    import pandas as pd

    workdir = built_blade["workdir"]
    mass_csv = workdir / "drape" / "test_blade_mass.csv"
    assert mass_csv.exists(), "Mass CSV file not found."
    mass_df = pd.read_csv(mass_csv, sep=",")

    ref_mass_csv = workdir.parent.parent / "data" / "test_blade_mass.csv"
    assert ref_mass_csv.exists(), "Reference mass CSV file not found."
    ref_mass_df = pd.read_csv(ref_mass_csv, sep=",")

    assert mass_df.round().equals(ref_mass_df.round()), "Mass tables do not match."


def test_build_geometry_only(temp_example_dir):
    """Test the geometry subcommand independently."""
    # original_dir = os.getcwd()
    # os.chdir(temp_example_dir)
    path = temp_example_dir / "blade_test.yml"
    assert path.exists(), f"YAML file {path} does not exist."
    try:
        state = AppState()
        build_app = BuildApp(state, Path(temp_example_dir / "blade_test.yml"))
        build_app.geometry()

        workdir = temp_example_dir / "temp_blade" / "mesh"
        assert workdir.exists(), "Working directory not created."
        assert (workdir / "test_blade.vtp").exists(), "Geometry file not created."
        assert (
            temp_example_dir / "temp_blade" / "test_blade_portable.yml"
        ).exists(), "Portable YAML not created."
    finally:
        pass
        # os.chdir(original_dir)



import vtk
import logging

logger = logging.getLogger(__name__)


def intp_c(x, points, const=2, clamp=True):
    try:
        sc = vtk.vtkCardinalSpline()
        sc.SetLeftConstraint(const)
        sc.SetRightConstraint(const)
        if clamp:
            sc.ClampValueOn()
        else:
            sc.ClampValueOff()
        for i in points:
            sc.AddPoint(i[0], i[1])

        o = [sc.Evaluate(i) for i in x]
        return x, o

    except Exception as e:
        logger.error(f"Error occurred while interpolating points: {points}")
        logger.exception(e)



import numpy as np
import pyvista as pv
import logging
import os

logger = logging.getLogger(__name__)


def failure_calc(
    STRESS,
    UPSTRAN,
    Xt,
    Xc,
    Yt,
    Yc,
    Zt,
    Zc,
    S12,
    S13,
    S23,
    THETAF,
    MGF,
    ANU12,
    ANU12f,
    E11,
    E11f,
):
    """Vectorized Puck failure criterion for multiple stress states and fracture plane angles."""
    STRESS = np.atleast_2d(STRESS)
    n = STRESS.shape[0]
    e = np.zeros((n, 2))  # [FF_index, IFF_index]
    THETAMAX = np.zeros(n)

    Pi = np.pi
    deg2rad = Pi / 180.0

    # Fiber Failure
    A = STRESS[:, 0] - (ANU12 - ANU12f * MGF * (E11 / E11f)) * (
        STRESS[:, 1] + STRESS[:, 2]
    )
    e[:, 0] = np.where(A >= 0, A / Xt, np.abs(A) / Xc)

    # Inter-Fiber Failure
    Rn = Yt
    Rn1 = S12
    Rnt = Yc / (2.0 * np.tan(deg2rad * THETAF))
    Pnt = -1.0 / (2.0 * np.tan(2.0 * deg2rad * THETAF))
    Pn1 = Pnt * (Rn1 / Rnt)

    J = np.arange(181)
    THETA = -Pi / 2.0 + J * deg2rad
    cos_theta = np.cos(THETA)
    sin_theta = np.sin(THETA)
    cos_theta2 = cos_theta**2
    sin_theta2 = sin_theta**2
    sin_cos_theta = sin_theta * cos_theta

    SFP = (
        STRESS[:, 1, np.newaxis] * cos_theta2
        + STRESS[:, 2, np.newaxis] * sin_theta2
        + 2.0 * STRESS[:, 5, np.newaxis] * sin_cos_theta
    )
    TNT = (
        -STRESS[:, 1, np.newaxis] * sin_cos_theta
        + STRESS[:, 2, np.newaxis] * sin_cos_theta
        + STRESS[:, 5, np.newaxis] * (cos_theta2 - sin_theta2)
    )
    TN1 = STRESS[:, 3, np.newaxis] * cos_theta + STRESS[:, 4, np.newaxis] * sin_theta

    IFF_tensile = np.sqrt(
        (SFP / Rn) ** 2
        + (TN1 / (Rn1 - Pn1 * SFP)) ** 2
        + (TNT / (Rnt - Pnt * SFP)) ** 2
    )
    IFF_compressive = np.sqrt(
        (TN1 / (Rn1 - Pn1 * SFP)) ** 2 + (TNT / (Rnt - Pnt * SFP)) ** 2
    )
    IFF = np.where(SFP >= 0, IFF_tensile, IFF_compressive)

    e[:, 1] = np.max(IFF, axis=1)
    max_indices = np.argmax(IFF, axis=1)
    THETAMAX = THETA[max_indices] / deg2rad

    return e, THETAMAX


def build_stiffness_matrix(E11, E22, E33, G12, G13, G23, nu12, nu13, nu23):
    """Build 6x6 stiffness matrix [C] for an orthotropic material in material coordinates."""
    S11 = 1 / E11
    S22 = 1 / E22
    S33 = 1 / E33
    S12 = -nu12 / E11
    S13 = -nu13 / E11
    S23 = -nu23 / E22
    S44 = 1 / G23
    S55 = 1 / G13
    S66 = 1 / G12

    S = np.array(
        [
            [S11, S12, S13, 0, 0, 0],
            [S12, S22, S23, 0, 0, 0],
            [S13, S23, S33, 0, 0, 0],
            [0, 0, 0, S44, 0, 0],
            [0, 0, 0, 0, S55, 0],
            [0, 0, 0, 0, 0, S66],
        ]
    )

    C = np.linalg.inv(S)
    return C


def strain_to_stress(strain, C):
    """Convert strain vector [εxx, εyy, εzz, εxy, εyz, εxz] to stress [σxx, σyy, σzz, σxy, σyz, σxz]."""
    strain_voigt = strain.copy()
    strain_voigt[:, 3:] *= 2
    stress = strain_voigt @ C.T
    return stress


def transform_stress_to_material(stress, theta):
    """Transform stress [σxx, σyy, σzz, σxy, σyz, σxz] to material coordinates at angle theta (degrees)."""
    c = np.cos(np.radians(theta))
    s = np.sin(np.radians(theta))
    c2 = c**2
    s2 = s**2
    cs = c * s

    T = np.array(
        [
            [c2, s2, 0, 2 * cs, 0, 0],
            [s2, c2, 0, -2 * cs, 0, 0],
            [0, 0, 1, 0, 0, 0],
            [-cs, cs, 0, c2 - s2, 0, 0],
            [0, 0, 0, 0, c, -s],
            [0, 0, 0, 0, s, c],
        ]
    )

    stress_material = stress @ T.T
    return stress_material


def get_ply_stack():
    """Return default ply stack configuration with angles relative to local_x."""
    return [
        {
            "theta": 0.0,  # 0° relative to local_x
            "material": {
                "E11": 150000.0,
                "E22": 10000.0,
                "E33": 10000.0,
                "G12": 5000.0,
                "G13": 5000.0,
                "G23": 4000.0,
                "nu12": 0.3,
                "nu13": 0.3,
                "nu23": 0.4,
                "Xt": 2000.0,
                "Xc": 1500.0,
                "Yt": 50.0,
                "Yc": 200.0,
                "Zt": 50.0,
                "Zc": 200.0,
                "S12": 80.0,
                "S13": 80.0,
                "S23": 80.0,
                "THETAF": 53.0,
                "MGF": 0.5,
                "ANU12": 0.3,
                "ANU12f": 0.25,
                "E11_puck": 150000.0,
                "E11f": 200000.0,
            },
        },
        {
            "theta": 45.0,  # 45° relative to local_x
            "material": {
                "E11": 150000.0,
                "E22": 10000.0,
                "E33": 10000.0,
                "G12": 5000.0,
                "G13": 5000.0,
                "G23": 4000.0,
                "nu12": 0.3,
                "nu13": 0.3,
                "nu23": 0.4,
                "Xt": 2000.0,
                "Xc": 1500.0,
                "Yt": 50.0,
                "Yc": 200.0,
                "Zt": 50.0,
                "Zc": 200.0,
                "S12": 80.0,
                "S13": 80.0,
                "S23": 80.0,
                "THETAF": 53.0,
                "MGF": 0.5,
                "ANU12": 0.3,
                "ANU12f": 0.25,
                "E11_puck": 150000.0,
                "E11f": 200000.0,
            },
        },
        {
            "theta": 90.0,  # 90° relative to local_x
            "material": {
                "E11": 120000.0,
                "E22": 8000.0,
                "E33": 8000.0,
                "G12": 4000.0,
                "G13": 4000.0,
                "G23": 3000.0,
                "nu12": 0.28,
                "nu13": 0.28,
                "nu23": 0.35,
                "Xt": 1800.0,
                "Xc": 1300.0,
                "Yt": 40.0,
                "Yc": 180.0,
                "Zt": 40.0,
                "Zc": 180.0,
                "S12": 70.0,
                "S13": 70.0,
                "S23": 70.0,
                "THETAF": 50.0,
                "MGF": 0.4,
                "ANU12": 0.28,
                "ANU12f": 0.23,
                "E11_puck": 120000.0,
                "E11f": 180000.0,
            },
        },
    ]


def compute_laminate_failure(mesh, ply_stack, alignment_threshold=0.9):
    """Compute Puck failure indices for a surface mesh with cell-based strains and local ply angles."""
    # Ensure cell normals and strains
    mesh = mesh.compute_normals(point_normals=False, cell_normals=True)
    strainid = "TOSTRAIN_1.000"
    if strainid not in mesh.point_data:
        raise ValueError(
            f"Mesh must contain point data array '{strainid}' with 6 components"
        )

    # Interpolate nodal strains to cell centers
    mesh = mesh.point_data_to_cell_data()
    strains = mesh.cell_data[strainid]
    n_cells = strains.shape[0]

    # Compute local x and y vectors
    normals = mesh.cell_data["Normals"]
    global_z = np.array([0, 0, 1])
    dot_z_n = np.sum(normals * global_z, axis=1, keepdims=True)
    local_x = global_z - dot_z_n * normals
    local_x_norm = np.linalg.norm(local_x, axis=1, keepdims=True)
    local_x = np.where(local_x_norm > 1e-10, local_x / local_x_norm, 0)
    local_y = np.cross(normals, local_x)
    local_y_norm = np.linalg.norm(local_y, axis=1, keepdims=True)
    local_y = np.where(local_y_norm > 1e-10, local_y / local_y_norm, 0)

    # Filter cells by alignment of local_x with global z
    dot_products = np.sum(local_x * global_z, axis=1)
    keep_mask = dot_products <= alignment_threshold

    if not np.any(keep_mask):
        logger.info("No cells meet the alignment threshold. Output file not created.")
        return None

    indices_to_keep = np.where(keep_mask)[0]
    new_mesh = mesh.extract_cells(indices_to_keep)
    new_strains = new_mesh.cell_data[strainid]
    new_local_x = local_x[keep_mask]
    new_local_y = local_y[keep_mask]
    n_cells = new_strains.shape[0]

    failure_data = {}

    for ply_idx, ply in enumerate(ply_stack):
        theta_global = np.zeros(n_cells)
        for i in range(n_cells):
            # Define ply angle relative to local_x
            local_x_i = new_local_x[i]
            local_y_i = new_local_y[i]
            # Fiber direction: rotate local_x by theta degrees in local x-y plane
            theta_rad = np.radians(ply["theta"])
            fiber_dir = np.cos(theta_rad) * local_x_i + np.sin(theta_rad) * local_y_i
            # Compute global angle of fiber direction relative to global x-axis
            theta_global[i] = np.degrees(np.arctan2(fiber_dir[1], fiber_dir[0]))

        mat = ply["material"]
        E11, E22, E33 = mat["E11"], mat["E22"], mat["E33"]
        G12, G13, G23 = mat["G12"], mat["G13"], mat["G23"]
        nu12, nu13, nu23 = mat["nu12"], mat["nu13"], mat["nu23"]
        Xt, Xc = mat["Xt"], mat["Xc"]
        Yt, Yc = mat["Yt"], mat["Yc"]
        Zt, Zc = mat["Zt"], mat["Zc"]
        S12, S13, S23 = mat["S12"], mat["S13"], mat["S23"]
        THETAF = mat["THETAF"]
        MGF = mat["MGF"]
        ANU12, ANU12f = mat["ANU12"], mat["ANU12f"]
        E11_puck, E11f = mat["E11_puck"], mat["E11f"]

        # Compute stresses for each cell
        stresses_material = np.zeros((n_cells, 6))
        for i in range(n_cells):
            C = build_stiffness_matrix(E11, E22, E33, G12, G13, G23, nu12, nu13, nu23)
            stress_global = strain_to_stress(new_strains[i : i + 1], C)
            stresses_material[i] = transform_stress_to_material(
                stress_global, theta_global[i]
            )

        UPSTRAN = np.zeros_like(stresses_material)
        e, THETAMAX = failure_calc(
            stresses_material,
            UPSTRAN,
            Xt,
            Xc,
            Yt,
            Yc,
            Zt,
            Zc,
            S12,
            S13,
            S23,
            THETAF,
            MGF,
            ANU12,
            ANU12f,
            E11_puck,
            E11f,
        )

        failure_data[f"FF_ply{ply_idx + 1}"] = e[:, 0]
        failure_data[f"IFF_ply{ply_idx + 1}"] = e[:, 1]
        failure_data[f"THETAMAX_ply{ply_idx + 1}"] = THETAMAX

        logger.info(f"Ply {ply_idx + 1} (θ={ply['theta']}° relative to local_x):")
        logger.info(f"  Max FF Index: {np.max(e[:, 0]):.3f}")
        logger.info(f"  Max IFF Index: {np.max(e[:, 1]):.3f}")
        logger.info(f"  Failure Cells: {np.sum(np.any(e >= 1.0, axis=1))}/{n_cells}")

    # Add local vectors to mesh
    new_mesh.cell_data["Local_X"] = new_local_x
    new_mesh.cell_data["Local_Y"] = new_local_y

    return new_mesh, failure_data


def compute_failure_for_meshes(mesh_files):
    """Process multiple mesh files and compute failure indices for each."""
    for mesh_file in mesh_files:
        try:
            mesh = pv.read(mesh_file).extract_surface()
            ply_stack = get_ply_stack()
            result = compute_laminate_failure(mesh, ply_stack, 0.9)

            if result is None:
                logger.info(f"No valid cells for mesh {mesh_file}; skipping.")
                continue

            new_mesh, failure_data = result
            for key, data in failure_data.items():
                new_mesh.cell_data[key] = data
            output_file = os.path.splitext(mesh_file)[0] + "_fail.vtu"
            new_mesh.save(output_file)
            logger.info(f"Written output to {output_file}")
        except Exception as e:
            logger.error(f"Error processing {mesh_file}: {e}")



#!/usr/bin/env python

from b3p.cli import yml_portable
import argparse
import numpy as np
import os
import pandas as pd
from matplotlib import pyplot as plt
from scipy.optimize import fmin
from ccblade.ccblade import CCAirfoil, CCBlade
from pathlib import Path
import math
import logging

logger = logging.getLogger(__name__)


def load_polar(pname: Path):
    """Load and interpolate a polar by name to a set alpha range."""
    logger.info(f"loading polar {pname}")
    if not os.path.isfile(pname):
        raise IOError(f"Polar {pname} not found")

    with open(pname, "r") as f:
        lines = f.readlines()

    read_start_index = next(
        i for i, line in enumerate(lines) if line.startswith("-180")
    )
    read_end_index = next((i for i, line in enumerate(lines) if "EOT" in line), None)

    lst = [
        [float(j) for j in line.split()]
        for line in lines[read_start_index:read_end_index]
    ]
    g = list(zip(*lst))
    alpha_new = (
        list(np.linspace(-180, -20, 25))
        + list(np.linspace(-19, 20, 29))
        + list(np.linspace(21, 180, 25))
    )
    cl = np.interp(alpha_new, g[0], g[1])
    cd = np.interp(alpha_new, g[0], g[2])
    cm = np.interp(alpha_new, g[0], g[3])
    return [alpha_new, cl, cd, cm]


def plot_interpolated_polars(t, data, of="polars.png"):
    """Plot interpolated polars with thickness values."""
    fig, ax = plt.subplots(3, 1, figsize=(12, 19))
    for n, i in enumerate(t):
        alpha, cl, cd, cm = data[:, :, n]
        ax[0].plot(alpha, cl, label=f"t={t[n]:.3f}")
        ax[1].plot(alpha, cd)
        ax[2].plot(cd, cl)
        ax[0].set_ylabel("Cl")
        ax[1].set_ylabel("Cd")
        ax[2].set_ylabel("Cm")
        ax[0].legend()
    fig.savefig(of)


def plot_polars(polars, of="polars_in.png"):
    """Plot polar data from a list."""
    fig, ax = plt.subplots(3, 1, figsize=(12, 16))
    for i in polars:
        alpha, cl, cd, cm = i[1]
        ax[0].plot(alpha, cl, label=i[0])
        ax[1].plot(alpha, cd)
        ax[2].plot(cd, cl)
        ax[0].set_ylabel("Cl")
        ax[1].set_ylabel("Cd")
        ax[2].set_ylabel("Cl - Cd")
        ax[0].legend()
        ax[0].grid()
        ax[1].grid()
        ax[2].grid()
    fig.tight_layout()
    fig.savefig(of)


def interpolate_polars(polars, tnew, of=None):
    """Interpolate polars to new thickness values and return CCAirfoil objects."""
    indices = range(len(polars[0][1][0]))
    t = [polar[0] for polar in polars]
    data = np.array(
        [
            [
                np.interp(
                    np.flip(tnew.values),
                    np.flip(t),
                    np.flip([polar[1][i][idx] for polar in polars]),
                )
                for idx in indices
            ]
            for i in range(4)
        ]
    )
    output_polars = [
        CCAirfoil(
            Re=[1e6],
            alpha=data[0, :, i],
            cl=data[1, :, i],
            cd=data[2, :, i],
            cm=data[3, :, i],
        )
        for i in reversed(range(len(tnew)))
    ]
    if of:
        plot_polars(polars, of.replace(".png", "_in.png"))
        plot_interpolated_polars(tnew, data, of)
    return output_polars


class RotorOptimizer:
    """Optimize rotor performance."""

    def __init__(self, rotor, uinf, rated_power=20e7, omega=None, maxiter=5):
        """Initialize the rotor optimizer with given parameters."""
        self.rotor = rotor
        self.uinf = uinf
        self.omega = omega
        self.rated_power = rated_power
        self.cache = {}
        self.latest = None
        self.maxiter = maxiter

    def evaluate(self, x, coefficients=False):
        """Evaluate rotor performance for given variables."""
        if self.omega is None:
            omega, pitch = x
        else:
            omega, pitch = self.omega, x[0]

        if (omega, pitch) not in self.cache:
            outputs, _ = self.rotor.evaluate(
                self.uinf, omega, pitch, coefficients=coefficients
            )
            self.cache[(omega, pitch)] = outputs

        P = np.mean(self.cache[(omega, pitch)]["P"])
        return P, self.cache[(omega, pitch)]

    def objective(self, x):
        """Compute objective function for optimization."""
        ev, _ = self.evaluate(x)
        return np.fabs(self.rated_power - ev)

    def optimize(self, initial_guess):
        """Optimize rotor variables using fmin."""
        result = fmin(self.objective, initial_guess, maxiter=self.maxiter)
        rr, rdet = self.evaluate(result)
        return result


def tsr2omega(tsr, uinf, radius, max_tip_speed=95.0):
    """Convert tip speed ratio to rotor speed in RPM."""
    ts = np.minimum(uinf * tsr, max_tip_speed)
    return (ts * 60.0) / (2.0 * np.pi * radius)


def omega2tsr(omega, uinf, radius):
    """Convert rotor speed to tip speed ratio."""
    return omega * 2.0 * np.pi * radius / (uinf * 60.0)


def plot_grid(num_plots, figsize=(15, 15)):
    """Create a grid of subplots for plotting."""
    grid_size = math.isqrt(num_plots)
    columns = grid_size
    rows = np.ceil(num_plots / grid_size).astype(int)
    fig, axs = plt.subplots(rows, columns, figsize=figsize)
    axs = axs.flatten()
    for idx in range(len(axs) - 1, rows * columns):
        fig.delaxes(axs[idx])
    return fig, axs


def find_closest_x(x_values, evaluations, target, order):
    """Find closest x value to target using polynomial interpolation."""
    assert len(x_values) == len(evaluations)
    poly_coeffs = np.polyfit(x_values, evaluations, order)
    poly = np.poly1d(poly_coeffs)
    x_dense = np.linspace(x_values[0], x_values[-1], 10000)
    x_closest = x_dense[np.argmin(np.abs(poly(x_dense) - target))]
    return x_closest


def plot_bladeloads(r, data_dict, of="bladeloads.png"):
    """Plot blade loads from a dictionary."""
    fig, axs = plot_grid(len(data_dict), figsize=(15, 15))
    for idx, (name, array) in enumerate(data_dict.items()):
        axs[idx].plot(r, array)
        axs[idx].set_title(name)
    fig.tight_layout()
    fig.savefig(of)
    logger.info(f"Saved {of}")


class controloptimize:
    """Optimize rotor control settings."""

    def __init__(self, rotor, max_tipspeed, rtip, rating, uinf, workdir):
        """Initialize control optimizer with rotor parameters."""
        self.rotor = rotor
        self.max_tipspeed = max_tipspeed
        self.rating = rating
        self.uinf = uinf
        self.rtip = rtip
        self.workdir = workdir

    def control_opt_below_rated(
        self, starting_uinf=6, starting_tsr=10, starting_pitch=0
    ):
        """Optimize control for below rated power at a specific wind speed."""
        omega = tsr2omega(
            starting_tsr, uinf=6, radius=self.rtip, max_tip_speed=self.max_tipspeed
        )
        optimizer = RotorOptimizer(self.rotor, [starting_uinf])
        initial_guess = np.array([omega, starting_pitch])
        optimal_values = optimizer.optimize(initial_guess)
        logger.info(f"initial guess {initial_guess} optimal values {optimal_values}")
        init_val = optimizer.evaluate(initial_guess)[0]
        opt_val, optt = optimizer.evaluate(optimal_values, coefficients=True)
        logger.info(f" {init_val} {opt_val}, improvement {opt_val / init_val}")
        self.optimal_tsr = omega2tsr(optimal_values[0], starting_uinf, self.rtip)
        self.fine_pitch = optimal_values[1]
        loads, _ = self.rotor.distributedAeroLoads(
            self.uinf, optimal_values[0], optimal_values[1], 0
        )
        plot_bladeloads(
            self.rotor.r, loads, of=os.path.join(self.workdir, "ccblade_bladeloads.png")
        )
        logger.info(f"optimal tsr {self.optimal_tsr} {self.fine_pitch}")

    def control_opt_above_rated(self):
        """Optimize control for above rated power by adjusting pitch."""
        self.omega = tsr2omega(
            self.optimal_tsr,
            uinf=self.uinf,
            radius=self.rtip,
            max_tip_speed=self.max_tipspeed,
        )
        self.tsr = omega2tsr(self.omega, uinf=self.uinf, radius=self.rtip)
        self.pitch = self.fine_pitch * np.ones_like(self.uinf)
        init_pc, _ = self.rotor.evaluate(
            self.uinf,
            self.omega,
            self.pitch,
            coefficients=False,
        )
        rotorplot(init_pc, self.uinf, of=os.path.join(self.workdir, "ccblade_init.png"))
        overrated = np.where(init_pc["P"] > self.rating)
        logger.info(f"overrated {overrated}, {self.uinf[overrated]}")
        upost = self.uinf[overrated]
        ompost = self.omega[overrated]
        pitch_over_rated = []
        closest_pitch = self.fine_pitch
        for i in np.array(list(zip(upost, ompost))):
            pitch = np.linspace(closest_pitch, closest_pitch + 15.0, 8)
            ui = np.ones_like(pitch) * i[0]
            oi = np.ones_like(pitch) * i[1]
            pg, _ = self.rotor.evaluate(ui, oi, pitch, coefficients=False)
            closest_pitch = find_closest_x(pitch, pg["P"], self.rating, 3)
            pitch_over_rated.append(closest_pitch)
        self.pitch[overrated] = pitch_over_rated
        out_pc, _ = self.rotor.evaluate(
            self.uinf,
            self.omega,
            self.pitch,
            coefficients=True,
        )
        out_pc["omega"] = self.omega
        out_pc["tsr"] = self.tsr
        out_pc["pitch"] = self.pitch
        out_pc["uinf"] = self.uinf
        rotorplot(
            out_pc,
            self.uinf,
            labels=["P", "CP", "Mb", "T", "omega", "pitch", "tsr"],
            of=os.path.join(self.workdir, "ccblade_out.png"),
        )
        logger.info(f"pitch {self.pitch}")
        return out_pc


def rotorplot(op, uinf, labels=["P", "CP", "T", "Mb"], of="__temp.png"):
    """Plot rotor performance data against wind speeds."""
    lab = [i for i in labels if i in op]
    fig, ax = plot_grid(len(lab), figsize=(10, 10))
    for n, i in enumerate(lab):
        ax[n].plot(uinf, op[i], label=f"{i} max={op[i].max():.2f}")
        ax[n].legend()
        ax[n].grid()
        ax[n].set_ylabel(i + "(W)" if i == "P" else "")
        ax[n].set_xlabel("uinf [m/s]")
    fig.tight_layout()
    fig.savefig(of)
    logger.info(f"saved {of}")


class ccblade_run:
    """Run CCBlade analysis on a blade."""

    def __init__(self, blade):
        """Initialize with a blade YAML file."""
        self.dct = yml_portable.yaml_make_portable(blade)
        workdir = os.path.join(self.dct["general"]["workdir"], "mesh")
        bem = {
            "rated_power": 10e6,
            "polars": {},
            "B": 3,
            "rho": 1.225,
            "tilt": 5,
            "precone": 3,
            "shearExp": 0.1,
            "hubHt": 140.0,
            "mu": 1.81206e-5,
            "yaw": 0,
            "max_tipspeed": 95.0,
            "uinf": [3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 16, 20],
        }
        missing_keys = [key for key in bem if key not in self.dct["aero"]["bem"]]
        logger.info(
            f"the following keys are not specified in aero/bem: {missing_keys}, using defaults {bem}"
        )
        bem |= self.dct["aero"]["bem"]
        self.prefix = os.path.join(workdir, self.dct["general"]["prefix"])
        if "polars" not in bem:
            exit("no polars in blade file")
        if not os.path.isfile(f"{self.prefix}.pck"):
            exit("blade not built yet, run b3p build <blade.yml> first")
        plf = pd.read_csv(f"{self.prefix}_sca_50.csv", sep=";")
        plrs = sorted(
            [(i[0], load_polar(Path(i[1]))) for i in bem["polars"].items()],
            reverse=True,
        )
        iplr = interpolate_polars(
            plrs, plf.relative_thickness, of=os.path.join(workdir, "polars.png")
        )
        rhub, rtip = plf.z.iloc[0], plf.z.iloc[-1]
        self.rotor = CCBlade(
            plf.z - plf.z[0],
            plf.chord,
            plf.twist,
            iplr,
            rhub,
            rtip,
            B=bem["B"],
            rho=bem["rho"],
            mu=bem["mu"],
            precone=bem["precone"],
            tilt=bem["tilt"],
            yaw=bem["yaw"],
            shearExp=bem["shearExp"],
            hubHt=bem["hubHt"],
            derivatives=False,
        )
        logger.info(f"Rotor from {rhub} to {rtip}")
        self.copt = controloptimize(
            self.rotor,
            bem["max_tipspeed"],
            rtip,
            bem["rated_power"],
            uinf=np.array(bem["uinf"]),
            workdir=workdir,
        )

    def run(self):
        """Execute the CCBlade analysis."""
        self.copt.control_opt_below_rated()
        output = self.copt.control_opt_above_rated()
        del output["W"]
        workdir = self.dct["general"]["workdir"]
        df = pd.DataFrame(output).dropna()
        df.to_csv(os.path.join(workdir, "ccblade_output.csv"), sep=";")


def main():
    """Run the main script with command-line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument("blade", help="blade file")
    args = parser.parse_args()
    ccblade_run(args.blade).run()


if __name__ == "__main__":
    main()



import logging
import vtk
import numpy as np
import pyvista as pv

logger = logging.getLogger(__name__)


def equals(v1, v2):
    tol = 1e-6
    return (v1 - v2) ** 2 < tol


def mesh_line(pnt1, pnt2, n_cells, id):
    xyz = []
    web_height = vtk.vtkGeoMath().DistanceSquared(pnt1, pnt2) ** 0.5

    for i in zip(pnt1, pnt2):
        mm = min(0.3, 0.06 / web_height)
        rel = sorted([0, 1] + list(np.linspace(mm, 1.0 - mm, n_cells - 2)))
        ab = [j * (i[1] - i[0]) + i[0] for j in rel]
        xyz.append(np.array(ab))

    dst = [i[1:] - i[:-1] for i in xyz]
    sl = (dst[0] ** 2 + dst[1] ** 2 + dst[2] ** 2) ** 0.5

    pl = [0] + [sum(sl[:i]) for i in range(1, len(sl) + 1)]

    ppl = [-i + pl[-1] for i in pl]
    ml = [abs(i - 0.5 * web_height) for i in pl]

    wh = [web_height for _ in ml]

    rad = np.mean(xyz[2])
    r = [rad for _ in range(n_cells)]

    arrays = {
        "d_te": pl,
        "d_le": ppl,
        "d_le_r": [i / max(ppl) for i in ppl],
        f"d_{id}_r": [i / max(ml) for i in ml],
        f"d_{id}": ml,
        "web_height": wh,
        "radius": r,
        "is_web": [1.0 for _ in ppl],
    }

    return list(zip(*xyz)), arrays


class web:
    def __init__(
        self, points, web_root, web_tip, web_name, coordinate, flip_normal=False
    ):
        self.points = points
        self.web_root = web_root
        self.web_tip = web_tip
        spl1, spl2 = vtk.vtkSCurveSpline(), vtk.vtkSCurveSpline()
        for i in points:
            spl1.AddPoint(i[0], i[1])
            spl2.AddPoint(i[0], i[2])

        self.splines = (spl1, spl2)
        self.evaluations = {}
        self.name = web_name
        self.coordinate = coordinate
        self.flip_normal = flip_normal

    def average_splits(self):
        g = list(zip(*self.points))
        return np.mean(g[1]), np.mean(g[2])

    def splits(self, r, r_relative):
        out = (0, 0)
        out = (self.splines[0].Evaluate(r), self.splines[1].Evaluate(r))
        self.evaluations[int(round(r * 1e2) * 10)] = [out]
        return out

    def _find_top_and_bottom_points(self, mesh):
        rad = mesh.GetPointData().GetArray("radius")
        rel_dist = mesh.GetPointData().GetArray("d_rel_dist_from_te")
        for i in range(mesh.GetNumberOfPoints()):
            rm = rad.GetValue(i)
            if self.web_root <= rm <= self.web_tip:
                rd = rel_dist.GetValue(i)
                pnt = mesh.GetPoint(i)
                rmm = int(round(rm * 1e2) * 10)
                if equals(rd, self.evaluations[rmm][0][0]) or equals(
                    rd, self.evaluations[rmm][0][1]
                ):
                    self.evaluations[rmm].append(pnt)

    def _create_quad_connectivity(self, n_points, n_total, flip=False):
        nrows = int(n_total / n_points)
        colids = np.arange(n_points - 1)
        npp = (
            np.arange(1, nrows).repeat(n_points - 1).reshape(nrows - 1, n_points - 1)
            - 1
        ) * n_points + colids
        if flip:
            stck = [
                np.ones_like(npp) * 4,
                npp + 1,
                npp + n_points + 1,
                npp + n_points,
                npp,
            ]
        else:
            stck = [
                np.ones_like(npp) * 4,
                npp,
                npp + n_points,
                npp + n_points + 1,
                npp + 1,
            ]
        return np.stack(stck).T.flatten()

    def _create_points(self, n_cells):
        ev = self.evaluations
        vp = []
        added_arrays = {}
        for i in sorted(ev):
            if len(ev[i]) == 3:
                pnts, data = mesh_line(ev[i][1], ev[i][2], n_cells, self.coordinate)
                vp.extend(pnts)
                for j in data:
                    if j not in added_arrays:
                        added_arrays[j] = data[j]
                    else:
                        added_arrays[j].extend(data[j])

        return vp, added_arrays

    def write_mesh(self, vtpfile):
        self.mesh.save(vtpfile)
        logger.info(f"Wrote mesh to {vtpfile}")

    def mesh(self, mesh, n_cells):
        self._find_top_and_bottom_points(mesh)
        points, pdata = self._create_points(n_cells)

        cells = self._create_quad_connectivity(n_cells, len(points), self.flip_normal)

        self.mesh = pv.PolyData(points, faces=cells)

        for i in pdata:
            self.mesh.point_data[i] = np.array(pdata[i]).astype(np.float32)



import subprocess
from pathlib import Path


def build_ccblade_dependency():
    repo_name = "ccblade"
    build_dir = Path("build") / repo_name
    if not build_dir.exists():
        subprocess.run(
            ["git", "clone", "https://github.com/WISDEM/CCBlade.git", build_dir],
            check=True,
        )
    subprocess.run(["pip", "install", "."], cwd=build_dir, check=True)


if __name__ == "__main__":
    build_ccblade_dependency()



import pytest
from pathlib import Path
import glob
import os
from b3p.cli.app_state import AppState
from b3p.cli.ccx_app import CcxApp
from b3p.cli.build_app import BuildApp
import filecmp
import logging

logger = logging.getLogger(__name__)


@pytest.fixture(scope="session")
def run_ccx(temp_example_dir):
    """Fixture to run the CCX process after a build."""
    # original_dir = os.getcwd()
    # os.chdir(temp_example_dir)
    try:
        state = AppState()
        yml_path = Path(temp_example_dir / "blade_test.yml")

        build_app = BuildApp(state, yml_path)
        ccx_app = CcxApp(state, yml_path)
        build_app.build()
        ccx_app.prep(bondline=False)
        yield {
            "workdir": temp_example_dir / "temp_blade",
            "yml_path": yml_path,
            "ccx_app": ccx_app,
            "temp_dir": temp_example_dir.parent,
        }
    finally:
        pass
        # os.chdir(original_dir)


def test_ccx_prep(run_ccx):
    """Test if CCX preparation generates the expected input file."""
    workdir = run_ccx["workdir"]

    logger.info(f"Checking CCX prep in workdir: {workdir}")
    inp_files = glob.glob(f"{workdir}/fea/*_ccx_*.inp")
    assert inp_files, "CCX prep should generate at least one .inp file"
    assert os.path.exists(
        inp_files[0]
    ), f"Expected CCX input file {inp_files[0]} not found"


def test_ccx_bondline_selection(run_ccx):
    """Test if bondline mesh is selected when bondline=True."""
    workdir = run_ccx["workdir"]
    bondline_vtu = glob.glob(f"{workdir}/drape/*_bondline.vtu")
    assert bondline_vtu, "Bondline VTU should exist from the build process"
    inp_files = glob.glob(f"{workdir}/fea/*_ccx_*.inp")
    assert inp_files[0].startswith(
        str(workdir / "fea" / "test_blade")
    ), "CCX input file should be generated from bondline mesh"


def test_ccx_produce_fwd_edge_inp(run_ccx):
    """Test if CCX produces the forward edge input file."""
    workdir = run_ccx["workdir"]
    inp_files = glob.glob(f"{workdir}/fea/*_ccx_*.inp")
    assert inp_files, "CCX prep should generate at least one .inp file"
    assert any(
        "_forward_edge" in f for f in inp_files
    ), "CCX prep should produce a forward edge input file"


@pytest.mark.skip(
    reason="Seems a element numbering issue on the web, not sure if this gives different results, skip for now"
)
def test_ccx_forward_edge_content(run_ccx):
    """Test if the generated forward edge .inp file matches the reference file."""
    workdir = run_ccx["workdir"]
    temp_dir = run_ccx["temp_dir"]
    edgewise_loadcase = workdir / "fea" / "test_blade_ccx_lc_forward_edge.inp"
    assert edgewise_loadcase.exists(), "CCX prep should generate a forward edge input file named *_ccx_lc_forward_edge.inp"
    # generated_file = generated_files[0]

    reference_file = temp_dir / "data" / "test_blade_ccx_lc_forward_edge.inp"
    assert (
        reference_file.exists()
    ), f"Reference file {reference_file} not found in temp data directory"

    cmp = filecmp.cmp(edgewise_loadcase, reference_file, shallow=False)
    assert cmp, f"Generated file {edgewise_loadcase} does not match reference file {reference_file}"



#! /usr/bin/env python3

from b3p.mesh import mesh_from_loft
from b3p.mesh import webs
import numpy as np


def build_blade_structure(config, prefix=None):
    pckfile = f"{prefix}.pck"
    base_vtp = f"{prefix}_base.vtp"
    shell_vtp = f"{prefix}_shell.vtp"

    radii = eval(str(config["mesh"]["radii"]))
    config_webs = config["mesh"]["webs"]
    panel_mesh_scale = (
        config["mesh"]["panel_mesh_scale"]
        if "panel_mesh_scale" in config["mesh"]
        else []
    )

    # build the plain blade mesh
    mesh_from_loft.build_mesh(pckfile, radii, [], [], prefix, outfile=base_vtp)

    # compute the locations where the webs intersect with the blade mesh
    web_shell_intersections = webs.build_webs(base_vtp, config_webs, prefix=prefix)

    # rejoggle the datums format from the yaml file into dict
    if "coordinates" in config["mesh"]:
        ad = config["mesh"]["coordinates"]
        added_datums = dict(
            [(i, [ad[i]["base"]] + list(zip(*ad[i]["points"]))) for i in ad]
        )
    else:
        added_datums = {}

    # build the blade mesh with webs
    return mesh_from_loft.build_mesh(
        pckfile,
        radii,
        config_webs,
        web_shell_intersections,
        prefix,
        config["mesh"]["n_web_points"],
        config["mesh"]["n_chordwise_points"],
        outfile=shell_vtp,
        added_datums=added_datums,
        panel_mesh_scale=panel_mesh_scale,
    )



import logging
from pathlib import Path
import os
import shutil

logger = logging.getLogger(__name__)


class CleanApp:
    def __init__(self, state, yml: Path):
        self.state = state
        self.yml = yml

    def clean(self):
        dct = self.state.load_yaml(self.yml)
        prefix = dct["general"]["workdir"]
        if os.path.isdir(prefix):
            shutil.rmtree(prefix)
            logger.info(f"Removed workdir {prefix}")
        else:
            logger.info(f"Workdir {prefix} does not exist")



#! /usr/bin/env python3

import vtk
import os
import multiprocessing
import argparse
import numpy as np
import pandas as pd
import math
import copy
import json
import pyvista as pv
import logging

logger = logging.getLogger(__name__)


def get_eids(sec, n):
    """Get cell IDs connected to a point."""
    g = vtk.vtkIdList()
    sec.GetPointCells(n, g)
    return [g.GetId(i) for i in range(g.GetNumberOfIds())]


def get_nids(sec, e):
    """Get point IDs for a given cell."""
    l = vtk.vtkIdList()
    sec.GetCellPoints(e, l)
    return [l.GetId(i) for i in range(l.GetNumberOfIds())]


def join_up(nodes, epid, stacks, sec, poly):
    """Link nodes at ply drops using Delaunay triangulation with scaled thicknesses."""
    cells = poly.GetPolys()
    mat = poly.GetCellData().GetArray("mat")
    angle = poly.GetCellData().GetArray("angle")
    for i, el in nodes:
        s0, s1 = (
            np.array(stacks[el[0]], dtype=float),
            np.array(stacks[el[1]], dtype=float),
        )
        s0 /= np.linalg.norm(s0)
        s1 /= np.linalg.norm(s1)

        pnts = vtk.vtkPoints()
        for j in s0:
            pnts.InsertNextPoint(0, j, 0)
        for j in s1:
            pnts.InsertNextPoint(5, j, 0)

        pl = vtk.vtkPolyData()
        pl.SetPoints(pnts)
        d2 = vtk.vtkDelaunay2D()
        d2.SetInputData(pl)
        d2.Update()
        dd = d2.GetOutput()

        x0 = 1 if get_nids(sec, el[0])[0] != get_nids(sec, el[1])[1] else 0
        nmap = list(range(epid[el[0]] + x0, epid[el[0]] + x0 + len(s0) * 2, 2)) + list(
            range(epid[el[1]] + 1 - x0, epid[el[1]] + 1 - x0 + len(s1) * 2, 2)
        )
        for j in range(dd.GetNumberOfCells()):
            c = dd.GetCell(j)
            cells.InsertNextCell(4)
            for k in range(c.GetNumberOfPoints()):
                cells.InsertCellPoint(nmap[c.GetPointId(k)])
            cells.InsertCellPoint(nmap[c.GetPointId(k)])
            mat.InsertNextTuple1(-1)
            angle.InsertNextTuple1(0)


def align_normals(align):
    """Reverse cells with normals pointing in the positive z-direction."""
    for i in range(align.GetNumberOfCells()):
        c = align.GetCell(i)
        p = c.GetPoints()
        if p.GetNumberOfPoints() > 2:
            normal = [0, 0, 0]
            vtk.vtkTriangle().ComputeNormal(
                p.GetPoint(0), p.GetPoint(1), p.GetPoint(2), normal
            )
            if normal[2] > 0:
                align.ReverseCell(i)


def web_link(web_links, epid, stacks, sec, poly):
    """Link web elements to shell elements."""
    enormal = sec.GetCellData().GetArray("Normals")
    mat = poly.GetCellData().GetArray("mat")
    angle = poly.GetCellData().GetArray("angle")
    cells = poly.GetPolys()

    for i in web_links:
        to1, to2 = i[2][0], i[2][1]
        s1, s2, s0 = stacks[to1], stacks[to2], stacks[i[0]]
        n11, n12 = epid[to1] + 2 * len(s1) - 1, epid[to1] + 2 * len(s1) - 2
        n21, n22 = epid[to2] + 2 * len(s2) - 1, epid[to2] + 2 * len(s2) - 2
        n1, n2 = enormal.GetTuple3(i[0]), enormal.GetTuple3(to2)

        linkto = (
            [j + epid[to2] for j in range(0, len(s2) * 2, 2)]
            if max(n1[0] * n2[0], n1[1] * n2[1]) > 0.5
            else [n11, n12, n21, n22]
        )
        nds = linkto + [j + epid[i[0]] + i[3] for j in range(0, len(s0) * 2, 2)]

        pts = vtk.vtkPoints()
        p1, p2 = poly.GetPoint(n11), poly.GetPoint(nds[4])
        dx = -1 if linkto[0] == n11 and p1[0] < p2[0] else 1 if linkto[0] == n11 else 0

        for j in nds[:4]:
            p = poly.GetPoint(j)
            pts.InsertNextPoint(p[0] + dx, p[1], 0)
        for j in nds[4:]:
            p = poly.GetPoint(j)
            pts.InsertNextPoint(p[0], p[1], 0)

        pol = vtk.vtkPolyData()
        pol.SetPoints(pts)
        d2 = vtk.vtkDelaunay2D()
        d2.SetInputData(pol)
        d2.Update()
        dd = d2.GetOutput()

        for j in range(dd.GetNumberOfCells()):
            c = dd.GetCell(j)
            cells.InsertNextCell(4)
            for k in range(c.GetNumberOfPoints()):
                cells.InsertCellPoint(nds[c.GetPointId(k)])
            cells.InsertCellPoint(nds[c.GetPointId(k)])
            mat.InsertNextTuple1(-1)
            angle.InsertNextTuple1(0)


def create_bondline(y, epid, stacks, sec, poly, bondline_material=20):
    """Create a bondline at the trailing edge."""
    lst = []
    bnds = sec.GetBounds()
    for i in range(sec.GetNumberOfPoints()):
        if sec.GetPoint(i)[1] > bnds[3] - y:
            lst.extend(get_eids(sec, i))

    pts = vtk.vtkPoints()
    nds = []
    for j in set(lst):
        s = stacks[j]
        n1, n2 = epid[j] + 2 * len(s) - 1, epid[j] + 2 * len(s) - 2
        pts.InsertNextPoint(poly.GetPoint(n1))
        pts.InsertNextPoint(poly.GetPoint(n2))
        nds.extend((n1, n2))

    pol = vtk.vtkPolyData()
    pol.SetPoints(pts)
    d2 = vtk.vtkDelaunay2D()
    d2.SetInputData(pol)
    dd = d2.GetOutput()

    cells = poly.GetPolys()
    material = poly.GetCellData().GetArray("mat")
    angle = poly.GetCellData().GetArray("angle")
    for j in range(dd.GetNumberOfCells()):
        c = dd.GetCell(j)
        cells.InsertNextCell(4)
        for k in range(c.GetNumberOfPoints()):
            cells.InsertCellPoint(nds[c.GetPointId(k)])
        cells.InsertCellPoint(nds[c.GetPointId(k)])
        material.InsertNextTuple1(bondline_material)
        angle.InsertNextTuple1(0)


def get_local_twist(r, var):
    return np.interp(r, var["z"][1] - np.min(var["z"][1]), var["twist"][1])


def get_local_chord(r, var):
    return np.interp(r, var["z"][1] - np.min(var["z"][1]), var["chord"][1])


def get_local_thickness(r, var):
    return np.interp(r, var["z"][1] - np.min(var["z"][1]), var["thickness"][1])


def write_vtp(section, vtp):
    """Write a VTK PolyData object to a .vtp file."""
    wr = vtk.vtkXMLPolyDataWriter()
    wr.SetFileName(vtp)
    wr.SetInputData(section)
    wr.Update()
    wr.Write()
    logger.info(f"writing to {vtp}")


def get_avg_normal(normals):
    mean_normal = np.mean(normals, axis=0)
    mean_normal /= np.linalg.norm(mean_normal)
    return mean_normal


def get_node_avg_normal(connected_cell_normals):
    s = connected_cell_normals.shape
    if s[0] < 3:
        return get_avg_normal(connected_cell_normals), None
    corr = np.zeros((s[0], s[0]))
    for i in range(s[0]):
        for j in range(i + 1, s[0]):
            corr[i, j] = np.abs(
                np.dot(connected_cell_normals[i, :], connected_cell_normals[j, :])
            )
    if corr[0, 1] > np.mean([corr[0, 2], corr[1, 2]]):
        return get_avg_normal(connected_cell_normals[:2, :]), 2
    elif corr[0, 2] > np.mean([corr[0, 1], corr[1, 2]]):
        return get_avg_normal(connected_cell_normals[::2, :]), 1
    else:
        return get_avg_normal(connected_cell_normals[1:, :]), 0


def slice_mesh(vtu, r, rotz):
    """Slice the 3D mesh at radius r and apply initial rotation."""
    rd = pv.read(vtu)  # Using PyVista for reading, then converting to VTK
    sec = rd.slice(normal=[0, 0, 1], origin=[0, 0, r])
    transform = vtk.vtkTransform()
    transform.RotateZ(rotz)
    transformfilter = vtk.vtkTransformFilter()
    transformfilter.SetTransform(transform)
    transformfilter.SetInputData(sec)
    transformfilter.Update()
    return transformfilter.GetOutput()


def compute_section_properties(sec, r, var):
    """Compute mid-position and geometric properties of the section."""
    pts = np.array([sec.GetPoint(i) for i in range(sec.GetNumberOfPoints())])
    mid_position = 0.5 * (np.amin(pts, axis=0) + np.amax(pts, axis=0))
    bnds = sec.GetBounds()
    return pd.DataFrame(
        [
            [r]
            + list(mid_position)
            + [
                get_local_twist(r, var),
                get_local_chord(r, var),
                get_local_thickness(r, var),
            ]
            + [bnds[1] - bnds[0], bnds[3] - bnds[2]]
        ],
        columns=[
            "r",
            "xavg",
            "yavg",
            "zavg",
            "twist_angle",
            "local_chord",
            "local_thickness",
            "dx",
            "dy",
        ],
    ), mid_position


def link_unconnected_points(sec):
    """Link unconnected points by adjusting their positions."""
    for i in range(sec.GetNumberOfPoints()):
        pb = sec.GetPoint(i)
        e0 = get_eids(sec, i)
        if len(e0) == 1:
            for j in np.linspace(0, 1, 100):
                p = sec.FindPoint(pb[0], pb[1] - j, pb[2])
                if p != i:
                    sec.GetPoints().SetPoint(i, sec.GetPoint(p))
                    break
    cln = vtk.vtkCleanPolyData()
    cln.SetInputData(sec)
    cln.Update()
    return cln.GetOutput()


def build_2d_mesh(sec):
    """Build the 2D mesh with ply stacks."""
    poly = vtk.vtkPolyData()
    points = vtk.vtkPoints()
    cells = vtk.vtkCellArray()
    poly.SetPoints(points)
    poly.SetPolys(cells)

    material = vtk.vtkIntArray()
    material.SetName("mat")
    poly.GetCellData().AddArray(material)
    angle = vtk.vtkFloatArray()
    angle.SetName("angle")
    poly.GetCellData().AddArray(angle)

    nply = sec.GetCellData().GetArray("n_plies")
    thickness = sec.GetCellData().GetArray("thickness")
    enormal = sec.GetCellData().GetArray("Normals")
    plies = [
        (sec.GetCellData().GetArrayName(i), sec.GetCellData().GetArray(i))
        for i in range(sec.GetCellData().GetNumberOfArrays())
        if sec.GetCellData().GetArrayName(i).startswith("ply")
    ]

    pid = 0
    epid = []
    join_nodes = []
    web_links = []
    stacks, mstacks, astacks = [], [], []

    for i in range(sec.GetNumberOfCells()):
        epid.append(pid)
        nn = get_nids(sec, i)
        e0, e1 = get_eids(sec, nn[0]), get_eids(sec, nn[1])
        e0b, e1b = copy.deepcopy(e0), copy.deepcopy(e1)

        nr0, wn0 = get_node_avg_normal(np.array([enormal.GetTuple3(e) for e in e0]))
        nr1, wn1 = get_node_avg_normal(np.array([enormal.GetTuple3(e) for e in e1]))

        if wn0 is not None:
            del e0[wn0]
        if wn1 is not None:
            del e1[wn1]

        x0, x1 = 0.0, 1.0
        if nply.GetTuple1(e0[0]) != nply.GetTuple1(e0[1]) or thickness.GetTuple1(
            e0[0]
        ) != thickness.GetTuple1(e0[1]):
            join_nodes.append((nn[0], e0))
            x0 = 0.1
        if nply.GetTuple1(e1[0]) != nply.GetTuple1(e1[1]) or thickness.GetTuple1(
            e1[0]
        ) != thickness.GetTuple1(e1[1]):
            x1 = 0.9

        if wn0 is not None and i == e0b[wn0]:
            x0, x1 = 0, 1.0
            fact = 1.1
            cell_length = sec.GetCell(i).GetLength2() ** 0.5
            nr0 = np.array(enormal.GetTuple3(i))
            nr0[2] = 0.0
            nr0 /= np.linalg.norm(nr0)
            web_links.append([i, nn[0], e0b, 0])
            th = thickness.GetTuple1(e0b[0 if wn0 != 0 else 1]) * 1e-3
            x0 = min(fact * th / cell_length, 0.97)

        if wn1 is not None and i == e1b[wn1]:
            x0, x1 = 0, 1.0
            fact = 1.1
            cell_length = sec.GetCell(i).GetLength2() ** 0.5
            nr1 = np.array(enormal.GetTuple3(i))
            nr1[2] = 0.0
            nr1 /= np.linalg.norm(nr1)
            web_links.append([i, nn[1], e1b, 1])
            th = thickness.GetTuple1(e1[0]) * 1e-3
            x1 = max(1.0 - fact * th / cell_length, 0.03)

        p0 = tuple(
            (1.0 - x0) * a + x0 * b
            for a, b in zip(sec.GetPoint(nn[0]), sec.GetPoint(nn[1]))
        )
        p1 = tuple(
            (1.0 - x1) * a + x1 * b
            for a, b in zip(sec.GetPoint(nn[0]), sec.GetPoint(nn[1]))
        )
        points.InsertNextPoint(p0)
        points.InsertNextPoint(p1)
        pid += 2

        stack, mstack, astack = [0], [], []
        for _, ply_data in plies:
            tup = ply_data.GetTuple(i)
            if tup[0] != -1 and tup[1] > 0.01:
                if mstack and mstack[-1] == int(tup[0]) and astack[-1] == tup[2]:
                    stack[-1] += tup[1]
                    p0 = tuple(k[0] + k[1] * tup[1] * 1e-3 for k in zip(p0, nr0))
                    p1 = tuple(k[0] + k[1] * tup[1] * 1e-3 for k in zip(p1, nr1))
                    npts = points.GetNumberOfPoints()
                    points.SetPoint(npts - 2, p0)
                    points.SetPoint(npts - 1, p1)
                else:
                    stack.append(tup[1] + stack[-1])
                    mstack.append(int(tup[0]))
                    astack.append(tup[2])
                    p0 = tuple(k[0] + k[1] * tup[1] * 1e-3 for k in zip(p0, nr0))
                    p1 = tuple(k[0] + k[1] * tup[1] * 1e-3 for k in zip(p1, nr1))
                    points.InsertNextPoint(p0)
                    points.InsertNextPoint(p1)
                    pid += 2

        stacks.append(stack)
        mstacks.append(mstack)
        astacks.append(astack)

    for i in range(sec.GetNumberOfCells()):
        p1, p2 = epid[i], epid[i] + 1
        for j in range(len(mstacks[i])):
            cells.InsertNextCell(4)
            cells.InsertCellPoint(p1 + 2 * j)
            cells.InsertCellPoint(p1 + 2 * j + 2)
            cells.InsertCellPoint(p2 + 2 * j + 2)
            cells.InsertCellPoint(p2 + 2 * j)
            material.InsertNextTuple1(mstacks[i][j])
            angle.InsertNextTuple1(astacks[i][j])

    return poly, epid, stacks, join_nodes, web_links


# def apply_transforms(poly, rotz, mid_position, local_twist):
#     """Apply transformations to align the 2D mesh in a single step."""
#     cln = vtk.vtkCleanPolyData()
#     cln.SetInputData(poly)
#     cln.Update()

#     transform = vtk.vtkTransform()
#     transform.RotateZ(rotz)
#     transform.Translate(-mid_position[0], -mid_position[1], -mid_position[2])
#     transform.RotateZ(local_twist)
#     transformfilter = vtk.vtkTransformFilter()
#     transformfilter.SetTransform(transform)
#     transformfilter.SetInputData(cln.GetOutput())
#     transformfilter.Update()
#     return transformfilter.GetOutput()


# def apply_transforms(poly, rotz, mid_position, local_twist):
#     """Apply transformations to align the 2D mesh in a single step."""
#     cln = vtk.vtkCleanPolyData()
#     cln.SetInputData(poly)
#     cln.Update()

#     transform = vtk.vtkTransform()
#     transform.PreMultiply()  # Explicitly set to ensure matrix composition order
#     transform.RotateZ(rotz)
#     transform.Translate(-mid_position[0], -mid_position[1], -mid_position[2])
#     transform.RotateZ(local_twist)
#     transformfilter = vtk.vtkTransformFilter()
#     transformfilter.SetTransform(transform)
#     transformfilter.SetInputData(cln.GetOutput())
#     transformfilter.Update()
#     return transformfilter.GetOutput()


def apply_transforms(poly, rotz, mid_position, local_twist):
    """Apply transformations to align the 2D mesh."""
    cln = vtk.vtkCleanPolyData()
    cln.SetInputData(poly)
    cln.Update()

    transform = vtk.vtkTransform()
    transform.RotateZ(rotz)
    transformfilter = vtk.vtkTransformFilter()
    transformfilter.SetTransform(transform)
    transformfilter.SetInputData(cln.GetOutput())
    transformfilter.Update()

    transform = vtk.vtkTransform()
    transform.Translate(-mid_position[0], -mid_position[1], -mid_position[2])
    transformfilter2 = vtk.vtkTransformFilter()
    transformfilter2.SetTransform(transform)
    transformfilter2.SetInputData(transformfilter.GetOutput())
    transformfilter2.Update()

    transform = vtk.vtkTransform()
    transform.RotateZ(local_twist)
    transformfilter3 = vtk.vtkTransformFilter()
    transformfilter3.SetTransform(transform)
    transformfilter3.SetInputData(transformfilter2.GetOutput())
    transformfilter3.Update()
    return transformfilter3.GetOutput()


def assign_triangle_properties(poly):
    """Assign material and angle properties to triangle elements."""
    mat = poly.GetCellData().GetArray("mat")
    ang = poly.GetCellData().GetArray("angle")
    for i in range(poly.GetNumberOfCells()):
        c = poly.GetCell(i)
        if c.GetNumberOfPoints() == 3 and mat.GetTuple1(i) == -1:
            g = c.GetPointIds()
            nb = []
            for j in [(0, 1), (1, 2), (2, 0)]:
                t = vtk.vtkIdList()
                t.InsertNextId(g.GetId(j[0]))
                t.InsertNextId(g.GetId(j[1]))
                idl = vtk.vtkIdList()
                poly.GetCellNeighbors(i, t, idl)
                nb.extend(idl.GetId(k) for k in range(idl.GetNumberOfIds()))
            mat.SetComponent(i, 0, -1)
            ang.SetComponent(i, 0, 0)
            for j in nb:
                if poly.GetCell(j).GetNumberOfPoints() == 4:
                    mat.SetComponent(i, 0, mat.GetTuple1(j))
                    ang.SetComponent(i, 0, ang.GetTuple1(j))


def compute_angle2(poly):
    """Compute and add angle2 array based on cell orientations."""
    ang2 = vtk.vtkFloatArray()
    ang2.SetName("angle2")
    ang2.SetNumberOfTuples(poly.GetNumberOfCells())
    angle_lookup = {}

    for i in range(poly.GetNumberOfCells()):
        cl = poly.GetCell(i)
        if cl.GetNumberOfPoints() > 2:
            pt1, pt2 = [cl.GetEdge(1).GetPoints().GetPoint(j) for j in range(2)]
            vec = np.array([pt2[j] - pt1[j] for j in range(3)])
            vec /= np.linalg.norm(vec)
            dot = vtk.vtkMath.Dot([1, 0, 0], vec)
            angle = math.degrees(math.acos(dot)) * (-1 if pt2[1] < pt1[1] else 1)

            if cl.GetNumberOfPoints() == 3:
                for jj in range(3):
                    if cl.GetPointId(jj) in angle_lookup:
                        angle = angle_lookup[cl.GetPointId(jj)]
                        break
            else:
                angle_lookup[cl.GetPointId(0)] = angle
            ang2.SetComponent(i, 0, angle)

    poly.GetCellData().AddArray(ang2)


def cut_blade(r, vtu, if_bondline=True, rotz=0, var=None, verbose=False):
    """Create a 2D cross-sectional mesh from a 3D VTU file at radius r."""
    var = var or {}
    workdir = os.path.dirname(vtu)
    if not os.path.exists(os.path.join(workdir, "2d")):
        os.makedirs(os.path.join(workdir, "2d"), exist_ok=True)

    output_file = os.path.join(workdir, "2d", f"msec_{int(1e3 * r)}.vtp")

    if os.path.exists(output_file):
        logger.info(f"2d mesh {output_file} already exists - skipping ")
        return output_file

    logger.info(f"# creating cross section mesh from {vtu} at r={r:.3f}")
    sec = slice_mesh(vtu, r, rotz)
    table_out, mid_position = compute_section_properties(sec, r, var)

    ctp = vtk.vtkCellDataToPointData()
    ctp.PassCellDataOn()
    ctp.SetInputData(sec)
    ctp.Update()
    sec = ctp.GetOutput()

    sec = link_unconnected_points(sec)
    poly, epid, stacks, join_nodes, web_links = build_2d_mesh(sec)

    join_up(join_nodes, epid, stacks, sec, poly)
    web_link(web_links, epid, stacks, sec, poly)
    if if_bondline:
        create_bondline(0.1, epid, stacks, sec, poly)

    if verbose:
        write_vtp(poly, os.path.join(workdir, f"inter_{int(1e3 * r)}.vtp"))

    out = apply_transforms(poly, rotz, mid_position, get_local_twist(r, var))
    assign_triangle_properties(out)
    compute_angle2(out)

    if verbose:
        write_vtp(out, os.path.join(workdir, f"prerealign_{int(1e3 * r)}.vtp"))

    align_normals(out)
    write_vtp(out, output_file)
    table_out.to_csv(
        os.path.join(workdir, "2d", f"section_location_{int(1e3 * r)}.csv"), index=False
    )
    logger.info(f"# written vtk to {output_file}")
    return output_file


def run_parallel_tasks(func, items, args_list):
    """Execute a function over a list of items in parallel or sequentially."""
    if len(items) == 0:
        return []
    parallel = args_list[-1]  # Last argument is parallel flag
    if parallel:
        with multiprocessing.Pool() as pool:
            results = pool.starmap(func, [(item, *args_list[:-1]) for item in items])
        return results
    return [func(item, *args_list[:-1]) for item in items]


def cut_blade_parallel(vtu, rr, if_bondline, rotz, var, verbose=False, parallel=False):
    """Process multiple radii in parallel or sequentially."""
    var_data = json.load(open(var, "r"))
    args = [vtu, if_bondline, rotz, var_data, verbose]
    return run_parallel_tasks(cut_blade, rr, args + [parallel])


def main():
    """Command-line interface for slicing 3D meshes into 2D sections."""
    p = argparse.ArgumentParser(
        description="Slice through a 3D FEA mesh (joined) to create 2D solid meshes for computing aeroelastic beam properties"
    )
    p.add_argument("--vtu", required=True, help="Input VTU mesh file")
    p.add_argument("--r", default="[30]", help="Radius array (e.g., '[10, 20, 30]')")
    p.add_argument("--bondline", action="store_true", help="Include bondline")
    p.add_argument("--rotz", type=float, default=0, help="Z-axis rotation angle")
    p.add_argument("--var", type=str, required=True, help="Path to variable file")
    p.add_argument("--verbose", action="store_true", help="Enable verbose output")
    p.add_argument("--parallel", action="store_true", help="Run in parallel")
    args = p.parse_args()

    rr = eval(args.r)
    cut_blade_parallel(
        args.vtu,
        rr,
        args.bondline,
        args.rotz,
        args.var,
        args.verbose,
        args.parallel,
    )


if __name__ == "__main__":
    main()



#! /usr/bin/env python3

import os
from copy import deepcopy as dc
import b3p.geometry.blade
import b3p.geometry.loft_utils
import numpy as np


def build_blade_geometry(config, prefix, xfoil=True):
    """Perform blade 3d model generation based on b3p dictionary."""
    pln = config["planform"]
    blade = b3p.geometry.blade.blade(
        pln["chord"],
        pln["thickness"],
        pln["twist"],
        dc(pln["dx"]),
        pln["dy"],
        pln["z"],
        config["aero"]["airfoils"],
        chordwise_sampling=b3p.geometry.loft_utils.optspace(
            config["planform"]["npchord"]
        ),
        np_spanwise=config["planform"]["npspan"],
    )

    blade.mesh(f"{prefix}.vtp")
    blade.dump(f"{prefix}.pck", z_rotation=0)
    blade.export_variables(f"{prefix}_variables.json")
    if xfoil:
        blade.export_xfoil(
            prefix=os.path.join(
                prefix.parent,
                "airfoil_out",
                f"xs_{config['general']['prefix']}",
            )
        )
    blade.plot(f"{prefix}", fname=f"{prefix}.png")
    n_sections = 50
    blade.to_table(np.linspace(0, 1, n_sections), "%s_sca_%i" % (prefix, n_sections))

    return blade



import pyvista as pv
import numpy as np
import json
import pytest
import logging

logger = logging.getLogger(__name__)


def test_laminate_number_of_plies(built_blade):
    """Test if the number of plies is correct."""
    workdir = built_blade["workdir"]
    temp_dir = built_blade["temp_dir"]
    joined = workdir / "drape" / "test_blade_joined.vtu"
    # joined = glob.glob(f"{workdir}/drape/*_joined.vtu")[-1]
    vtu = pv.read(joined)
    cell_arrays = [i for i in vtu.cell_data.keys() if "ply" in i]
    assert len(cell_arrays) == 348, f"Expected 348 plies, got {len(cell_arrays)}"

    sums = [float(vtu.cell_data[i][:, 1].sum()) for i in cell_arrays]
    logger.info(f"Computed ply sums: {sums}")

    json_path = temp_dir / "data" / "laminate_sums.json"
    if not json_path.exists():
        pytest.skip("Reference JSON 'laminate_sums.json' not found in tests/data/")
    with open(json_path, "r") as f:
        expected_sums = json.load(f)

    np.testing.assert_allclose(
        sums, expected_sums, atol=1e-4, err_msg="Ply sums do not match expected values"
    )
